"""
.. _ref_operator:

Operator
========
"""

import logging
import os
import traceback
import warnings

from enum import Enum
from ansys.dpf.core.check_version import version_requires, server_meet_version
from ansys.dpf.core.config import Config
from ansys.dpf.core.errors import DpfVersionNotSupported
from ansys.dpf.core.inputs import Inputs
from ansys.dpf.core.mapping_types import types
from ansys.dpf.core.common import types_enum_to_types
from ansys.dpf.core.outputs import Output, Outputs, _Outputs
from ansys.dpf.core import server as server_module
from ansys.dpf.core.operator_specification import Specification
from ansys.dpf.core.unit_system import UnitSystem
from ansys.dpf.gate import (
    operator_capi,
    operator_abstract_api,
    operator_grpcapi,
    data_processing_capi,
    data_processing_grpcapi,
    collection_capi,
    collection_grpcapi,
    dpf_vector,
    object_handler,
)

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")


class _SubOperator:
    def __init__(self, op_name, op_to_connect):
        self.op_name = op_name
        self.op = Operator(self.op_name, server=op_to_connect._server)
        if op_to_connect.inputs is not None:
            for key in op_to_connect.inputs._connected_inputs:
                inpt = op_to_connect.inputs._connected_inputs[key]
                if type(inpt).__name__ == "dict":
                    for keyout in inpt:
                        if inpt[keyout]() is not None:
                            self.op.connect(key, inpt[keyout](), keyout)
                else:
                    self.op.connect(key, inpt())

    def __call__(self):
        return self.op


class Operator:
    """Represents an operator, which is an elementary operation.

    The operator is the only object used to create and transform
    data. When the operator is evaluated, it processes the
    input information to compute its output with respect to its
    description.

    Parameters
    ----------
    name : str
        Name of the operator. For example, ``"U"``. You can use the
        ``"html_doc"`` operator to retrieve a list of existing operators.

    config : Config, optional
        The Configuration allows to customize how the operation
        will be processed by the operator. The default is ``None``.

    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    Create an operator from the library of operators.

    >>> from ansys.dpf import core as dpf
    >>> disp_oper = dpf.operators.result.displacement()

    Create an operator from a model.

    >>> from ansys.dpf.core import Model
    >>> from ansys.dpf.core import examples
    >>> model = Model(examples.find_static_rst())
    >>> disp_oper = model.results.displacement()

    """

    def __init__(self, name, config=None, server=None):
        """Initialize the operator with its name by connecting to a stub."""
        self.name = name
        self._internal_obj = None
        self._description = None
        self._inputs = None

        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        # step 2: get api
        self._api_instance = None  # see _api property

        # step3: init environment
        self._api.init_operator_environment(self)  # creates stub when gRPC

        # step4: if object exists: take instance, else create it (server)
        if self._server.has_client():
            self._internal_obj = self._api.operator_new_on_client(self.name, self._server.client)
        else:
            self._internal_obj = self._api.operator_new(self.name)

        if self._internal_obj is None:
            raise KeyError(
                f"The operator {self.name} doesn't exist in the registry. "
                f"Check its spelling in the documentation or verify its availability "
                f"in your loaded plugins. The current available operator names can be "
                f"accessed using 'available_operator_names' method."
            )

        self._spec = Specification(operator_name=self.name, server=self._server)
        # add dynamic inputs
        if len(self._spec.inputs) > 0 and self._inputs is None:
            self._inputs = Inputs(self._spec.inputs, self)

        # step4: if object exists: take instance (config)
        if config:
            self.config = config

        self._description = self._spec.description
        self._progress_bar = False

    @property
    def _api(self) -> operator_abstract_api.OperatorAbstractAPI:
        if self._api_instance is None:
            self._api_instance = self._server.get_api_for_type(
                capi=operator_capi.OperatorCAPI,
                grpcapi=operator_grpcapi.OperatorGRPCAPI,
            )
        return self._api_instance

    def _add_sub_res_operators(self, sub_results):
        """Dynamically add operators for instantiating subresults.

        Subresults for new operators are connected to the parent
        operator's inputs when created but are then completely
        independent of them.

        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> model = Model(examples.find_static_rst())
        >>> disp_oper = model.results.displacement()
        >>> disp_oper = model.results.displacement()
        >>> disp_x = model.results.displacement().X()
        >>> disp_y = model.results.displacement().Y()
        >>> disp_z = model.results.displacement().Z()

        """

        for result_type in sub_results:
            try:
                setattr(
                    self,
                    result_type["name"],
                    _SubOperator(result_type["operator name"], self),
                )
            except KeyError:
                pass

    @property
    def _outputs(self):
        if self._spec and len(self._spec.outputs) != 0:
            return Outputs(self._spec.outputs, self)

    @_outputs.setter
    def _outputs(self, value):
        # the Operator should not hold a reference on its outputs because outputs hold a reference
        # on the Operator
        pass

    @property
    @version_requires("3.0")
    def progress_bar(self) -> bool:
        """With this property, the user can choose to print a progress bar when
        the operator's output is requested, default is False"""
        return self._progress_bar

    @progress_bar.setter
    def progress_bar(self, value: bool) -> None:
        self._progress_bar = value

    def connect(self, pin, inpt, pin_out=0):
        """Connect an input on the operator using a pin number.

        Parameters
        ----------
        pin : int
            Number of the input pin.

        inpt : str, int, double, bool, list[int], list[float], Field, FieldsContainer, Scoping,
        ScopingsContainer, MeshedRegion, MeshesContainer, DataSources, CyclicSupport, dict, Outputs
            Operator, os.PathLike Object to connect to.

        pin_out : int, optional
            If the input is an operator, the output pin of the input operator. The default is ``0``.

        Examples
        --------
        Compute the minimum of displacement by chaining the ``"U"`` and ``"min_max_fc"`` operators.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.find_multishells_rst())
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        >>> max_fc_op = dpf.operators.min_max.min_max_fc()
        >>> max_fc_op.inputs.connect(disp_op.outputs)
        >>> max_field = max_fc_op.outputs.field_max()
        >>> max_field.data
        DPFArray([[0.59428386, 0.00201751, 0.0006032 ]]...

        """
        if inpt is self:
            raise ValueError("Cannot connect to itself.")
        elif isinstance(inpt, Operator):
            self._api.operator_connect_operator_output(self, pin, inpt, pin_out)
        elif isinstance(inpt, Output):
            self._api.operator_connect_operator_output(self, pin, inpt._operator, inpt._pin)
        elif isinstance(inpt, list):
            from ansys.dpf.core import collection

            if server_meet_version("3.0", self._server):
                inpt = collection.Collection.integral_collection(inpt, self._server)
                self._api.operator_connect_collection_as_vector(self, pin, inpt)
            else:
                if all(isinstance(x, int) for x in inpt):
                    self._api.operator_connect_vector_int(self, pin, inpt, len(inpt))
                else:
                    self._api.operator_connect_vector_double(self, pin, inpt, len(inpt))
        elif isinstance(inpt, dict):
            from ansys.dpf.core import label_space

            label_space_to_con = label_space.LabelSpace(
                label_space=inpt, obj=self, server=self._server
            )
            self._api.operator_connect_label_space(self, pin, label_space_to_con)
        elif isinstance(inpt, UnitSystem):
            if inpt.ID != -2:  # Ansys UnitSystem
                self.connect(pin, inpt.ID)
            else:  # Custom UnitSystem
                self.connect(pin, inpt.unit_names)
        else:
            if isinstance(inpt, os.PathLike):
                inpt = str(inpt)
            for type_tuple in self._type_to_input_method:
                if isinstance(inpt, type_tuple[0]):
                    if len(type_tuple) == 3:
                        inpt = type_tuple[2](inpt)
                    return type_tuple[1](self, pin, inpt)
            errormsg = f"input type {inpt.__class__} cannot be connected"
            raise TypeError(errormsg)

    @version_requires("6.2")
    def connect_operator_as_input(self, pin, op):
        """Connects an operator as an input on a pin.
        Parameters
        ----------
        pin : int
            Number of the output pin. The default is ``0``.
        op : :class:`ansys.dpf.core.dpf_operator.Operator`
            Requested type of the output. The default is ``None``.
        """
        self._api.operator_connect_operator_as_input(self, pin, op)

    @property
    def _type_to_output_method(self):
        from ansys.dpf.core import (
            cyclic_support,
            data_sources,
            field,
            fields_container,
            meshed_region,
            meshes_container,
            property_field,
            string_field,
            custom_type_field,
            result_info,
            scoping,
            scopings_container,
            time_freq_support,
            data_tree,
            workflow,
            collection,
            streams_container,
            generic_data_container,
            mesh_info,
        )

        out = [
            (bool, self._api.operator_getoutput_bool),
            (int, self._api.operator_getoutput_int),
            (str, self._api.operator_getoutput_string),
            (float, self._api.operator_getoutput_double),
            (field.Field, self._api.operator_getoutput_field, "field"),
            (
                property_field.PropertyField,
                self._api.operator_getoutput_property_field,
                "property_field",
            ),
            (
                string_field.StringField,
                self._api.operator_getoutput_string_field,
                "string_field",
            ),
            (
                custom_type_field.CustomTypeField,
                self._api.operator_getoutput_custom_type_field,
                "field",
            ),
            (scoping.Scoping, self._api.operator_getoutput_scoping, "scoping"),
            (
                fields_container.FieldsContainer,
                self._api.operator_getoutput_fields_container,
                "fields_container",
            ),
            (
                scopings_container.ScopingsContainer,
                self._api.operator_getoutput_scopings_container,
                "scopings_container",
            ),
            (
                meshes_container.MeshesContainer,
                self._api.operator_getoutput_meshes_container,
                "meshes_container",
            ),
            (
                streams_container.StreamsContainer,
                self._api.operator_getoutput_streams,
                "streams_container",
            ),
            (
                data_sources.DataSources,
                self._api.operator_getoutput_data_sources,
                "data_sources",
            ),
            (
                cyclic_support.CyclicSupport,
                self._api.operator_getoutput_cyclic_support,
                "cyclic_support",
            ),
            (
                meshed_region.MeshedRegion,
                self._api.operator_getoutput_meshed_region,
                "mesh",
            ),
            (
                result_info.ResultInfo,
                self._api.operator_getoutput_result_info,
                "result_info",
            ),
            (
                time_freq_support.TimeFreqSupport,
                self._api.operator_getoutput_time_freq_support,
                "time_freq_support",
            ),
            (
                mesh_info.MeshInfo,
                "mesh_info",
            ),
            (workflow.Workflow, self._api.operator_getoutput_workflow, "workflow"),
            (data_tree.DataTree, self._api.operator_getoutput_data_tree, "data_tree"),
            (Operator, self._api.operator_getoutput_operator, "operator"),
            (
                dpf_vector.DPFVectorInt,
                self._api.operator_getoutput_int_collection,
                lambda obj: collection.IntCollection(
                    server=self._server, collection=obj
                ).get_integral_entries(),
            ),
            (
                dpf_vector.DPFVectorDouble,
                self._api.operator_getoutput_double_collection,
                lambda obj: collection.FloatCollection(
                    server=self._server, collection=obj
                ).get_integral_entries(),
            ),
        ]
        if hasattr(self._api, "operator_getoutput_generic_data_container"):
            out.append(
                (
                    generic_data_container.GenericDataContainer,
                    self._api.operator_getoutput_generic_data_container,
                    "generic_data_container",
                )
            )
        return out

    @property
    def _type_to_input_method(self):
        from ansys.dpf.core import (
            cyclic_support,
            data_sources,
            field,
            collection,
            meshed_region,
            property_field,
            string_field,
            custom_type_field,
            scoping,
            time_freq_support,
            data_tree,
            workflow,
            model,
            generic_data_container,
        )

        out = [
            (bool, self._api.operator_connect_bool),
            ((int, Enum), self._api.operator_connect_int),
            (str, self._api.operator_connect_string),
            (float, self._api.operator_connect_double),
            (field.Field, self._api.operator_connect_field),
            (property_field.PropertyField, self._api.operator_connect_property_field),
            (string_field.StringField, self._api.operator_connect_string_field),
            (
                custom_type_field.CustomTypeField,
                self._api.operator_connect_custom_type_field,
            ),
            (scoping.Scoping, self._api.operator_connect_scoping),
            (collection.Collection, self._api.operator_connect_collection),
            (data_sources.DataSources, self._api.operator_connect_data_sources),
            (
                model.Model,
                self._api.operator_connect_data_sources,
                lambda obj: obj.metadata.data_sources,
            ),
            (cyclic_support.CyclicSupport, self._api.operator_connect_cyclic_support),
            (meshed_region.MeshedRegion, self._api.operator_connect_meshed_region),
            # TO DO: (result_info.ResultInfo, self._api.operator_connect_result_info),
            (
                time_freq_support.TimeFreqSupport,
                self._api.operator_connect_time_freq_support,
            ),
            (workflow.Workflow, self._api.operator_connect_workflow),
            (data_tree.DataTree, self._api.operator_connect_data_tree),
            (Operator, self._api.operator_connect_operator_as_input),
        ]
        if hasattr(self._api, "operator_connect_generic_data_container"):
            out.append(
                (
                    generic_data_container.GenericDataContainer,
                    self._api.operator_connect_generic_data_container,
                )
            )
        return out

    def get_output(self, pin=0, output_type=None):
        """Retrieve the output of the operator on the pin number.

        To activate the progress bar for server version higher or equal to 3.0,
        use ``my_op.progress_bar=True``

        Parameters
        ----------
        pin : int, optional
            Number of the output pin. The default is ``0``.
        output_type : :class:`ansys.dpf.core.common.types`, type,  optional
            Requested type of the output. The default is ``None``.

        Returns
        -------
        type
            Output of the operator.
        """
        output_type = _write_output_type_to_type(output_type)
        if self._server.meet_version("3.0") and self.progress_bar:
            self._server.session.add_operator(self, pin, "operator")
            self._progress_thread = self._server.session.listen_to_progress()
        if output_type is None:
            return self._api.operator_run(self)
        out = None
        for type_tuple in self._type_to_output_method:
            if output_type is type_tuple[0]:
                if len(type_tuple) >= 3:
                    if isinstance(type_tuple[2], str):
                        parameters = {type_tuple[2]: type_tuple[1](self, pin)}
                        out = output_type(**parameters, server=self._server)
                    else:
                        out = type_tuple[2](type_tuple[1](self, pin))
                if out is None:
                    try:
                        return output_type(type_tuple[1](self, pin), server=self._server)
                    except TypeError:
                        self._progress_thread = None
                        return output_type(type_tuple[1](self, pin))
        if out is not None:
            self._progress_thread = None
            return out
        raise TypeError(f"{output_type} is not an implemented Operator's output")

    @property
    def config(self):
        """Copy of the operator's current configuration.

        You can modify the copy of the configuration and then use ``operator.config = new_config``
        or instantiate an operator with the new configuration as a parameter.

        For information on an operator's options, see the documentation for that operator.

        Returns
        ----------
        :class:`ansys.dpf.core.config.Config`
            Copy of the operator's current configuration.

        Examples
        --------
        Modify the copy of an operator's configuration and set it as current config
        of the operator.

        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.add()
        >>> config_add = op.config
        >>> config_add.set_work_by_index_option(True)
        >>> op.config = config_add

        """
        config = self._api.operator_get_config(self)
        return Config(config=config, server=self._server, spec=self._spec)

    @config.setter
    def config(self, value):
        """Change the configuration of the operator.

        If the operator is up to date, changing the configuration
        doesn't make it not up to date.

        Parameters
        ----------
        value : Config
        """
        self._api.operator_set_config(self, value)

    @property
    def inputs(self):
        """Inputs connected to the operator.

        Returns
        --------
        :class:`ansys.dpf.core.inputs`
            Inputs connected to the operator.

        Examples
        --------
        Use the displacement operator.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.find_multishells_rst())
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)

        """

        return self._inputs

    @property
    def outputs(self):
        """Outputs from the operator's evaluation.

        Returns
        --------
        :class:`ansys.dpf.core.outputs`
            Outputs from the operator's evaluation.

        Examples
        --------
        Use the displacement operator.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.find_multishells_rst())
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        >>> disp_fc = disp_op.outputs.fields_container()

        """
        return self._outputs

    @staticmethod
    def default_config(name, server=None):
        """Retrieve the default configuration for an operator.

        You can change the copy of the default configuration to meet your needs
        before instantiating the operator.
        The Configuration allows to customize how the operation
        will be processed by the operator.

        Parameters
        ----------
        name : str
            Name of the operator.  For example ``"U"``. You can use the
            ``"html_doc"`` operator to retrieve a list of existing operators.
        server : server.DPFServer, optional
            Server with the channel connected to the remote or local instance. The
            default is ``None``, in which case an attempt is made to use the global
            server.

        Returns
        -------
        :class"`ansys.dpf.core.config.Config`
            Default configuration for the operator.

        """
        return Config(operator_name=name, server=server)

    def __del__(self):
        try:
            if self._internal_obj is not None:
                self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        str
            Description of the entity.
        """
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    def run(self):
        """Evaluate this operator."""
        self.get_output()

    def eval(self, pin=None):
        """Evaluate this operator.

        Parameters
        ----------
        pin : int
            Number of the output pin. The default is ``None``.

        Returns
        -------
        output : FieldsContainer, Field, MeshedRegion, Scoping
            Returns the first output of the operator by default and the output of a
            given pin when specified. Or, it only evaluates the operator without output.

        Examples
        --------
        Use the ``eval`` method.

        >>> from ansys.dpf import core as dpf
        >>> import ansys.dpf.core.operators.math as math
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.find_multishells_rst())
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        >>> normfc = math.norm_fc(disp_op).eval()

        """

        if not pin:
            if self.outputs != None and len(self.outputs._outputs) > 0:
                return self.outputs._outputs[0]()
            else:
                self.run()
        else:
            for output in self.outputs._outputs:
                if output._pin == pin:
                    return output()

    def _find_outputs_corresponding_pins(self, type_names, inpt, pin, corresponding_pins):
        from ansys.dpf.core.results import Result

        for python_name in type_names:
            # appears to be an issue on Linux.  This check is here
            # because cpp mappings are a single type mapping and
            # sometimes the spec contains 'B' instead of 'bool'
            if python_name == "B":
                python_name = "bool"

            if type(inpt).__name__ == python_name:
                corresponding_pins.append(pin)
            elif isinstance(inpt, (_Outputs, Operator, Result)):
                if isinstance(inpt, Operator):
                    output_pin_available = inpt.outputs._get_given_output([python_name])
                elif isinstance(inpt, Result):
                    output_pin_available = inpt().outputs._get_given_output([python_name])
                else:
                    output_pin_available = inpt._get_given_output([python_name])
                for outputpin in output_pin_available:
                    corresponding_pins.append((pin, outputpin))
            elif isinstance(inpt, Output):
                for inpttype in inpt._python_expected_types:
                    if inpttype == python_name:
                        corresponding_pins.append(pin)
                if python_name == "Any":
                    corresponding_pins.append(pin)
            elif python_name == "Any":
                corresponding_pins.append(pin)

    def __add__(self, fields_b):
        """Add two fields or two fields containers.

        Returns
        -------
        add : operators.math.add_fc
        """
        from ansys.dpf.core import dpf_operator, operators

        if hasattr(operators, "math") and hasattr(operators.math, "add_fc"):
            op = operators.math.add_fc(self, fields_b, server=self._server)
        else:
            op = dpf_operator.Operator("add_fc", server=self._server)
            op.connect(0, self)
            op.connect(1, fields_b)
        return op

    def __sub__(self, fields_b):
        """Subtract two fields or two fields containers.

        Returns
        -------
        minus : operators.math.minus_fc
        """
        from ansys.dpf.core import dpf_operator, operators

        if hasattr(operators, "math") and hasattr(operators.math, "minus_fc"):
            op = operators.math.minus_fc(server=self._server)
        else:
            op = dpf_operator.Operator("minus_fc", server=self._server)
        op.connect(0, self)
        op.connect(1, fields_b)
        return op

    def __pow__(self, value):
        if value != 2:
            raise ValueError('Only the value "2" is supported.')
        from ansys.dpf.core import dpf_operator, operators

        if hasattr(operators, "math") and hasattr(operators.math, "sqr_fc"):
            op = operators.math.sqr_fc(server=self._server)
        else:
            op = dpf_operator.Operator("sqr_fc", server=self._server)
        op.connect(0, self)
        op.connect(1, value)
        return op

    def __mul__(self, value):
        """Multiply two fields or two fields containers.

        Returns
        -------
        mul : operators.math.generalized_inner_product_fc
        """
        from ansys.dpf.core import dpf_operator, operators

        if hasattr(operators, "math") and hasattr(operators.math, "generalized_inner_product_fc"):
            op = operators.math.generalized_inner_product_fc(server=self._server)
        else:
            op = dpf_operator.Operator("generalized_inner_product_fc", server=self._server)
        op.connect(0, self)
        op.connect(1, value)
        return op

    @staticmethod
    def operator_specification(op_name, server=None):
        """Documents an Operator with its description (what the Operator does),
        its inputs and outputs and some properties"""
        return Specification(operator_name=op_name, server=server)

    @property
    def specification(self):
        """Returns the Specification (or documentation) of this Operator

        Returns
        -------
        Specification
        """
        if isinstance(self._spec, Specification):
            return self._spec
        else:
            return Specification(operator_name=self.name, server=self._server)

    def __truediv__(self, inpt):
        if isinstance(inpt, Operator):
            op = Operator("div")
            op.connect(0, self, 0)
            op.connect(1, inpt, 0)
        elif isinstance(inpt, float):
            op = Operator("scale")
            op.connect(0, self, 0)
            op.connect(1, 1.0 / inpt)
        return op


def available_operator_names(server=None):
    """Returns the list of operator names available in the server.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
    list

    Notes
    -----
    Function available with server's version starting at 3.0. Not available for server
    of type GrpcServer.

    """
    if server is None:
        server = server_module._global_server()

    if not server.meet_version("3.0"):
        raise DpfVersionNotSupported("3.0")

    api = server.get_api_for_type(
        capi=data_processing_capi.DataProcessingCAPI,
        grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
    )
    api.init_data_processing_environment(server)  # creates stub when gRPC
    coll_api = server.get_api_for_type(
        capi=collection_capi.CollectionCAPI,
        grpcapi=collection_grpcapi.CollectionGRPCAPI,
    )
    coll_api.init_collection_environment(server)

    if server.has_client():
        coll_obj = object_handler.ObjHandler(
            data_processing_api=api,
            internal_obj=api.data_processing_list_operators_as_collection_on_client(server.client),
        )
    else:
        coll_obj = object_handler.ObjHandler(
            data_processing_api=api,
            internal_obj=api.data_processing_list_operators_as_collection(),
        )
    num = coll_api.collection_get_size(coll_obj)
    out = []
    for i in range(num):
        out.append(coll_api.collection_get_string_entry(coll_obj, i))
    return out


def _write_output_type_to_type(output_type):
    if isinstance(output_type, str):
        output_type = types[output_type]

    if isinstance(output_type, types):
        try:
            return types_enum_to_types()[output_type]
        except KeyError as e:
            raise TypeError(f"{output_type} is not an implemented Operator's output")
    return output_type



from .csv_to_field import csv_to_field 
from .data_tree_to_json import data_tree_to_json 
from .data_tree_to_txt import data_tree_to_txt 
from .deserializer import deserializer 
from .export_symbolic_workflow import export_symbolic_workflow 
from .field_to_csv import field_to_csv 
from .hdf5dpf_custom_read import hdf5dpf_custom_read 
from .hdf5dpf_generate_result_file import hdf5dpf_generate_result_file 
from .import_symbolic_workflow import import_symbolic_workflow 
from .json_to_data_tree import json_to_data_tree 
from .mechanical_csv_to_field import mechanical_csv_to_field 
from .migrate_file_to_vtk import migrate_file_to_vtk 
from .migrate_to_vtu import migrate_to_vtu 
from .serialize_to_hdf5 import serialize_to_hdf5 
from .serializer import serializer 
from .serializer_to_string import serializer_to_string 
from .string_deserializer import string_deserializer 
from .txt_to_data_tree import txt_to_data_tree 
from .vtk_export import vtk_export 
from .vtk_to_fields import vtk_to_fields 
from .vtu_export import vtu_export 
from .workflow_export_json import workflow_export_json 
from .workflow_import_json import workflow_import_json 



"""Build static source operators from DPF server."""
import copy
import os
from datetime import datetime
from textwrap import wrap

import black
import chevron
from ansys.dpf import core as dpf
from ansys.dpf.core import common
from ansys.dpf.core.dpf_operator import available_operator_names
from ansys.dpf.core.outputs import _make_printable_type
from ansys.dpf.core.mapping_types import map_types_to_python


def build_docstring(specification):
    """Used to generate class docstrings."""
    docstring = ""
    if specification.description:
        docstring += "\n".join(
            wrap(specification.description, subsequent_indent="    ")
        )
        docstring += "\n\n"
    docstring = docstring.rstrip()
    return docstring.replace('"', "'")


def map_types(cpp_types):
    """Map C++ object types to Python types."""
    types = []
    # These types don't get mapped to Python types
    types_to_ignore = ["vector", "umap", "enum"]
    for cpp_type in cpp_types:
        if any(type_name in cpp_type for type_name in types_to_ignore):
            continue
        else:
            types.append(map_types_to_python[cpp_type])
    return types


def update_type_names_for_ellipsis(type_names):
    # Remove vector and umap types from the Python type
    new_types = []
    for name in type_names:
        if name == "vector<double>" or name == "vector<int32>":
            new_types.append(name)
        elif "vector" not in name and "umap" not in name:
            new_types.append(name)
    return new_types


def build_pin_data(pins, output=False):
    """Build pin data for use within template."""
    pin_ids = [pin for pin in pins]
    pin_ids.sort()

    data = []
    for id in pin_ids:
        specification = pins[id]

        type_names = specification.type_names

        derived_class_type_name = specification.name_derived_class

        if specification.ellipsis:
            type_names = update_type_names_for_ellipsis(type_names)
        docstring_types = map_types(type_names)
        parameter_types = " or ".join(docstring_types)
        parameter_types = "\n".join(
            wrap(parameter_types, subsequent_indent="        ", width=60)
        )

        pin_name = specification.name
        pin_name = pin_name.replace("<", "_")
        pin_name = pin_name.replace(">", "_")

        main_type = docstring_types[0] if len(docstring_types) >= 1 else ""
        built_in_types = ("int", "double", "string", "bool", "float", "str", "dict")

        # Case where output pin has multiple types.
        multiple_types = len(type_names) >= 2
        printable_type_names = type_names
        if multiple_types and output:
            printable_type_names = [_make_printable_type(name) for name in type_names]

        pin_data = {
            "id": id,
            "name": pin_name,
            "pin_name": pin_name, # Base pin name, without numbers for when pin is ellipsis
            "has_types": len(type_names) >= 1,
            "has_derived_class": len(derived_class_type_name) >= 1,
            "multiple_types": multiple_types,
            "printable_type_names": printable_type_names,
            "types": type_names,
            "derived_type_name": derived_class_type_name,
            "types_for_docstring": parameter_types,
            "main_type": main_type,
            "built_in_main_type": main_type in built_in_types,
            "optional": specification.optional,
            "document": "\n".join(
                wrap(
                    specification.document.capitalize().lstrip(' '),
                    subsequent_indent="        ",
                    width=45,
                )
            ),
            "ellipsis": 0 if specification.ellipsis else -1,
        }

        if specification.ellipsis:
            # Create two pins for ellipsis field with exactly the same
            # properties, just different names, ids, and ellipsis values
            pin_data["name"] = pin_name + "1"
            data.append(pin_data)

            second_pin_data = copy.deepcopy(pin_data)
            second_pin_data["name"] = pin_name + "2"
            second_pin_data["id"] = id + 1
            second_pin_data["ellipsis"] = 1
            data.append(second_pin_data)
        else:
            data.append(pin_data)

    return data


def build_operator(
    specification, operator_name, class_name, capital_class_name, category
):

    input_pins = []
    if specification.inputs:
        input_pins = build_pin_data(specification.inputs)

    output_pins = []
    if specification.outputs:
        output_pins = build_pin_data(specification.outputs, output=True)
    multiple_output_types = any(pin["multiple_types"] for pin in output_pins)

    docstring = build_docstring(specification)

    specification_description = "\n".join(
        wrap(specification.description, subsequent_indent="            ")
    )

    date_and_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

    data = {
        "operator_name": operator_name,
        "class_name": class_name,
        "class_name_underlining": len(class_name)*"=",
        "capital_class_name": capital_class_name,
        "docstring": docstring,
        "specification_description": specification_description,
        "input_pins": input_pins,
        "output_pins": output_pins,
        "outputs": len(output_pins) >= 1,
        "multiple_output_types": multiple_output_types,
        "category": category,
        "date_and_time": date_and_time,
    }

    this_path = os.path.dirname(os.path.abspath(__file__))
    mustache_file = os.path.join(this_path, "operator.mustache")
    with open(mustache_file, "r") as f:
        cls = chevron.render(f, data)

    return black.format_str(cls, mode=black.FileMode())


def build_operators():
    print(f"Generating operators for server {dpf.SERVER.version}")

    this_path = os.path.dirname(os.path.abspath(__file__))

    available_operators = available_operator_names()

    print(f"{len(available_operators)} operators found to generate.")

    succeeded = 0
    done = 0
    for operator_name in available_operators:
        if succeeded == done + 100:
            done += 100
            print(f"{done} operators done...")
        specification = dpf.Operator.operator_specification(operator_name)

        category = specification.properties.get("category", "")
        if not category:
            raise ValueError(f"Category not defined for operator {operator_name}.")
        scripting_name = specification.properties.get("scripting_name", "")

        # Make directory for new category
        category_path = os.path.join(this_path, category)
        if not os.path.exists(category_path):
            os.mkdir(category_path)

        # Clean up scripting name
        if scripting_name == "":
            scripting_name = operator_name
        if "::" in scripting_name:
            scripting_name = scripting_name.split("::")[-1]
        if "." in scripting_name:
            scripting_name = scripting_name.split(".")[-1]

        # Get python class name from scripting name
        capital_class_name = common._snake_to_camel_case(scripting_name)

        # Write to operator file
        operator_file = os.path.join(category_path, scripting_name + ".py")
        with open(operator_file, "w") as f:
            try:
                operator_str = build_operator(
                    specification,
                    operator_name,
                    scripting_name,
                    capital_class_name,
                    category,
                )
                exec(operator_str, globals())
                f.write(operator_str)
                succeeded += 1
            except SyntaxError as e:
                error_message = (
                    f"Unable to generate {operator_name}, {scripting_name}, {capital_class_name}.\n"
                    f"Error message: {e}\n"
                )
                with open(os.path.join(this_path, "failures.txt"), "w") as error_file:
                    error_file.write(error_message)
                    error_file.write(f"Class: {operator_str}")
                print(error_message)

    print(f"Generated {succeeded} out of {len(available_operators)}")
    if succeeded == len(available_operators):
        print("Success")
        exit(0)
    else:
        print("Terminated with errors")
        exit(1)


if __name__ == "__main__":
    dpf.set_default_server_context(dpf.AvailableServerContexts.premium)
    dpf.start_local_server(config=dpf.AvailableServerConfigs.LegacyGrpcServer)
    build_operators()
    dpf.SERVER.shutdown()



"""
.. _ref_generic_data_container:

GenericDataContainer
====================
"""
import traceback
import warnings
import builtins

from ansys.dpf.core import server as server_module
from ansys.dpf.core import errors
from ansys.dpf.core.any import Any
from ansys.dpf.core import collection
from ansys.dpf.core.mapping_types import map_types_to_python


class GenericDataContainer:
    """Maps properties to their DPF supported Data Types.

    Parameters
    ----------
    generic_data_container : ctypes.c_void_p, ansys.grpc.dpf.generic_data_container_pb2.GenericDataContainer message, optional  # noqa: E501
    server : DPFServer, optional
        Server with channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Notes
    -----
    Class available with server's version starting at 7.0 (Ansys 2024 R1 pre0).
    """

    def __init__(self, generic_data_container=None, server=None):
        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        if not self._server.meet_version("7.0"):
            raise errors.DpfVersionNotSupported("7.0")

        # step 2: if object exists, take the instance, else create it
        self._api_instance = None
        self._api.init_generic_data_container_environment(self)  # creates stub when gRPC

        self._api.init_generic_data_container_environment(self)

        if generic_data_container is not None:
            self._internal_obj = generic_data_container
        else:
            if self._server.has_client():
                self._internal_obj = self._api.generic_data_container_new_on_client(
                    self._server.client
                )
            else:
                self._internal_obj = self._api.generic_data_container_new()

    @property
    def _api(self):
        from ansys.dpf.gate import (
            generic_data_container_capi,
            generic_data_container_grpcapi,
        )

        if not self._api_instance:
            self._api_instance = self._server.get_api_for_type(
                capi=generic_data_container_capi.GenericDataContainerCAPI,
                grpcapi=generic_data_container_grpcapi.GenericDataContainerGRPCAPI,
            )
            self._api.init_generic_data_container_environment(self)  # creates stub when gRPC

        return self._api_instance

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        description : str
            Description of the entity.
        """
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    def set_property(self, property_name, prop):
        """Register given property with the given name.

        Parameters
        ----------
        property_name : str
            Property name.
        prop : Int, String, Float, Field, StringField, GenericDataContainer, Scoping
            object instance.
        """

        any_dpf = Any.new_from(prop, self._server)
        self._api.generic_data_container_set_property_any(self, property_name, any_dpf)

    def get_property(self, property_name):
        """Get property with given name.

        Parameters
        ----------
        property_name : str
            Property name.

        Returns
        -------
        Property object instance.
        """
        any_ptr = self._api.generic_data_container_get_property_any(self, property_name)
        any_dpf = Any(any_ptr, self._server)
        output_type = self.get_property_description()[property_name]
        class_ = getattr(builtins, output_type, None)
        if class_ is None:
            from ansys.dpf import core

            class_ = getattr(core, output_type)

        return any_dpf.cast(class_)

    def get_property_description(self):
        """Get a dictionary description of properties by name and data type

        Returns
        -------
        description: dict
            Description of the GenericDataContainer's contents
        """

        coll_obj = collection.StringCollection(
            collection=self._api.generic_data_container_get_property_names(self),
            server=self._server,
        )
        property_names = coll_obj.get_integral_entries()

        coll_obj = collection.StringCollection(
            collection=self._api.generic_data_container_get_property_types(self),
            server=self._server,
        )
        property_types = coll_obj.get_integral_entries()

        python_property_types = []
        for _, property_type in enumerate(property_types):
            python_property_types.append(map_types_to_python[property_type])

        return dict(zip(property_names, python_property_types))

    def __del__(self):
        if self._internal_obj is not None:
            try:
                self._deleter_func[0](self._deleter_func[1](self))
            except Exception as e:
                print(str(e.args), str(self._deleter_func[0]))
                warnings.warn(traceback.format_exc())



from . import averaging
from . import filter
from . import geo
from . import invariant
from . import logic
from . import mapping
from . import math
from . import mesh
from . import metadata
from . import min_max
from . import result
from . import scoping
from . import serialization
from . import utility



"""
cgns_result_provider
====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class cgns_result_provider(Operator):
    """Read/compute names result from result streams.

    Parameters
    ----------
    time_scoping : Scoping, optional
        Time/freq (use doubles or field), time/freq
        set ids (use ints or scoping) or
        time/freq step ids (use scoping with
        timefreq_steps location) required in
        output
    mesh_scoping : Scoping or ScopingsContainer, optional
        Nodes or elements scoping required in output.
        the scoping's location indicates
        whether nodes or elements are asked.
        using scopings container enables to
        split the result fields container in
        domains
    streams_container : StreamsContainer, optional
        Result file container allowed to be kept open
        to cache data
    data_sources : DataSources
        Result file path container, used if no
        streams are set
    result_name : str, optional
        Name of the result to read. by default the
        name of the operator is taken.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.cgns_result_provider()

    >>> # Make input connections
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_result_name = str()
    >>> op.inputs.result_name.connect(my_result_name)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.cgns_result_provider(
    ...     time_scoping=my_time_scoping,
    ...     mesh_scoping=my_mesh_scoping,
    ...     streams_container=my_streams_container,
    ...     data_sources=my_data_sources,
    ...     result_name=my_result_name,
    ... )

    >>> # Get output data
    >>> result_fields = op.outputs.fields()
    """

    def __init__(
        self,
        time_scoping=None,
        mesh_scoping=None,
        streams_container=None,
        data_sources=None,
        result_name=None,
        config=None,
        server=None,
    ):
        super().__init__(
            name="cgns::cgns::result_provider", config=config, server=server
        )
        self._inputs = InputsCgnsResultProvider(self)
        self._outputs = OutputsCgnsResultProvider(self)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if streams_container is not None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if result_name is not None:
            self.inputs.result_name.connect(result_name)

    @staticmethod
    def _spec():
        description = """Read/compute names result from result streams."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_scoping",
                    type_names=["scoping", "vector<int32>"],
                    optional=True,
                    document="""Time/freq (use doubles or field), time/freq
        set ids (use ints or scoping) or
        time/freq step ids (use scoping with
        timefreq_steps location) required in
        output""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping", "scopings_container"],
                    optional=True,
                    document="""Nodes or elements scoping required in output.
        the scoping's location indicates
        whether nodes or elements are asked.
        using scopings container enables to
        split the result fields container in
        domains""",
                ),
                3: PinSpecification(
                    name="streams_container",
                    type_names=["streams_container"],
                    optional=True,
                    document="""Result file container allowed to be kept open
        to cache data""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=False,
                    document="""Result file path container, used if no
        streams are set""",
                ),
                17: PinSpecification(
                    name="result_name",
                    type_names=["string"],
                    optional=True,
                    document="""Name of the result to read. by default the
        name of the operator is taken.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Results""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(
            name="cgns::cgns::result_provider", server=server
        )

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsCgnsResultProvider
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsCgnsResultProvider
        """
        return super().outputs


class InputsCgnsResultProvider(_Inputs):
    """Intermediate class used to connect user inputs to
    cgns_result_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.cgns_result_provider()
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_result_name = str()
    >>> op.inputs.result_name.connect(my_result_name)
    """

    def __init__(self, op: Operator):
        super().__init__(cgns_result_provider._spec().inputs, op)
        self._time_scoping = Input(cgns_result_provider._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._time_scoping)
        self._mesh_scoping = Input(cgns_result_provider._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._streams_container = Input(
            cgns_result_provider._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams_container)
        self._data_sources = Input(cgns_result_provider._spec().input_pin(4), 4, op, -1)
        self._inputs.append(self._data_sources)
        self._result_name = Input(
            cgns_result_provider._spec().input_pin(17), 17, op, -1
        )
        self._inputs.append(self._result_name)

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        Time/freq (use doubles or field), time/freq
        set ids (use ints or scoping) or
        time/freq step ids (use scoping with
        timefreq_steps location) required in
        output

        Parameters
        ----------
        my_time_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Nodes or elements scoping required in output.
        the scoping's location indicates
        whether nodes or elements are asked.
        using scopings container enables to
        split the result fields container in
        domains

        Parameters
        ----------
        my_mesh_scoping : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator.

        Result file container allowed to be kept open
        to cache data

        Parameters
        ----------
        my_streams_container : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> # or
        >>> op.inputs.streams_container(my_streams_container)
        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Result file path container, used if no
        streams are set

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def result_name(self):
        """Allows to connect result_name input to the operator.

        Name of the result to read. by default the
        name of the operator is taken.

        Parameters
        ----------
        my_result_name : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> op.inputs.result_name.connect(my_result_name)
        >>> # or
        >>> op.inputs.result_name(my_result_name)
        """
        return self._result_name


class OutputsCgnsResultProvider(_Outputs):
    """Intermediate class used to get outputs from
    cgns_result_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.cgns_result_provider()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields = op.outputs.fields()
    """

    def __init__(self, op: Operator):
        super().__init__(cgns_result_provider._spec().outputs, op)
        self._fields = Output(cgns_result_provider._spec().output_pin(0), 0, op)
        self._outputs.append(self._fields)

    @property
    def fields(self):
        """Allows to get fields output of the operator

        Returns
        ----------
        my_fields : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.cgns_result_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields = op.outputs.fields()
        """  # noqa: E501
        return self._fields



"""
merge_meshes
============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class merge_meshes(Operator):
    """Take a set of mesh and assemble them in a unique one

    Parameters
    ----------
    meshes1 : MeshedRegion
        A vector of meshed region to merge or meshed
        region from pin 0 to ...
    meshes2 : MeshedRegion
        A vector of meshed region to merge or meshed
        region from pin 0 to ...
    merge_method : int, optional
        0: merge by distance, 1: merge by node id
        (default)
    box_size : float, optional
        Box size used when merging by distance.
        default value is 1e-10.
    remove_duplicate_elements : int, optional
        0: keep duplicate elements (default), 1:
        remove duplicate elements


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.merge_meshes()

    >>> # Make input connections
    >>> my_meshes1 = dpf.MeshedRegion()
    >>> op.inputs.meshes1.connect(my_meshes1)
    >>> my_meshes2 = dpf.MeshedRegion()
    >>> op.inputs.meshes2.connect(my_meshes2)
    >>> my_merge_method = int()
    >>> op.inputs.merge_method.connect(my_merge_method)
    >>> my_box_size = float()
    >>> op.inputs.box_size.connect(my_box_size)
    >>> my_remove_duplicate_elements = int()
    >>> op.inputs.remove_duplicate_elements.connect(my_remove_duplicate_elements)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.merge_meshes(
    ...     meshes1=my_meshes1,
    ...     meshes2=my_meshes2,
    ...     merge_method=my_merge_method,
    ...     box_size=my_box_size,
    ...     remove_duplicate_elements=my_remove_duplicate_elements,
    ... )

    >>> # Get output data
    >>> result_merges_mesh = op.outputs.merges_mesh()
    """

    def __init__(
        self,
        meshes1=None,
        meshes2=None,
        merge_method=None,
        box_size=None,
        remove_duplicate_elements=None,
        config=None,
        server=None,
    ):
        super().__init__(name="merge::mesh", config=config, server=server)
        self._inputs = InputsMergeMeshes(self)
        self._outputs = OutputsMergeMeshes(self)
        if meshes1 is not None:
            self.inputs.meshes1.connect(meshes1)
        if meshes2 is not None:
            self.inputs.meshes2.connect(meshes2)
        if merge_method is not None:
            self.inputs.merge_method.connect(merge_method)
        if box_size is not None:
            self.inputs.box_size.connect(box_size)
        if remove_duplicate_elements is not None:
            self.inputs.remove_duplicate_elements.connect(remove_duplicate_elements)

    @staticmethod
    def _spec():
        description = """Take a set of mesh and assemble them in a unique one"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="meshes",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""A vector of meshed region to merge or meshed
        region from pin 0 to ...""",
                ),
                1: PinSpecification(
                    name="meshes",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""A vector of meshed region to merge or meshed
        region from pin 0 to ...""",
                ),
                101: PinSpecification(
                    name="merge_method",
                    type_names=["int32"],
                    optional=True,
                    document="""0: merge by distance, 1: merge by node id
        (default)""",
                ),
                102: PinSpecification(
                    name="box_size",
                    type_names=["double"],
                    optional=True,
                    document="""Box size used when merging by distance.
        default value is 1e-10.""",
                ),
                103: PinSpecification(
                    name="remove_duplicate_elements",
                    type_names=["int32"],
                    optional=True,
                    document="""0: keep duplicate elements (default), 1:
        remove duplicate elements""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="merges_mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="merge::mesh", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMergeMeshes
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMergeMeshes
        """
        return super().outputs


class InputsMergeMeshes(_Inputs):
    """Intermediate class used to connect user inputs to
    merge_meshes operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_meshes()
    >>> my_meshes1 = dpf.MeshedRegion()
    >>> op.inputs.meshes1.connect(my_meshes1)
    >>> my_meshes2 = dpf.MeshedRegion()
    >>> op.inputs.meshes2.connect(my_meshes2)
    >>> my_merge_method = int()
    >>> op.inputs.merge_method.connect(my_merge_method)
    >>> my_box_size = float()
    >>> op.inputs.box_size.connect(my_box_size)
    >>> my_remove_duplicate_elements = int()
    >>> op.inputs.remove_duplicate_elements.connect(my_remove_duplicate_elements)
    """

    def __init__(self, op: Operator):
        super().__init__(merge_meshes._spec().inputs, op)
        self._meshes1 = Input(merge_meshes._spec().input_pin(0), 0, op, 0)
        self._inputs.append(self._meshes1)
        self._meshes2 = Input(merge_meshes._spec().input_pin(1), 1, op, 1)
        self._inputs.append(self._meshes2)
        self._merge_method = Input(merge_meshes._spec().input_pin(101), 101, op, -1)
        self._inputs.append(self._merge_method)
        self._box_size = Input(merge_meshes._spec().input_pin(102), 102, op, -1)
        self._inputs.append(self._box_size)
        self._remove_duplicate_elements = Input(
            merge_meshes._spec().input_pin(103), 103, op, -1
        )
        self._inputs.append(self._remove_duplicate_elements)

    @property
    def meshes1(self):
        """Allows to connect meshes1 input to the operator.

        A vector of meshed region to merge or meshed
        region from pin 0 to ...

        Parameters
        ----------
        my_meshes1 : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> op.inputs.meshes1.connect(my_meshes1)
        >>> # or
        >>> op.inputs.meshes1(my_meshes1)
        """
        return self._meshes1

    @property
    def meshes2(self):
        """Allows to connect meshes2 input to the operator.

        A vector of meshed region to merge or meshed
        region from pin 0 to ...

        Parameters
        ----------
        my_meshes2 : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> op.inputs.meshes2.connect(my_meshes2)
        >>> # or
        >>> op.inputs.meshes2(my_meshes2)
        """
        return self._meshes2

    @property
    def merge_method(self):
        """Allows to connect merge_method input to the operator.

        0: merge by distance, 1: merge by node id
        (default)

        Parameters
        ----------
        my_merge_method : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> op.inputs.merge_method.connect(my_merge_method)
        >>> # or
        >>> op.inputs.merge_method(my_merge_method)
        """
        return self._merge_method

    @property
    def box_size(self):
        """Allows to connect box_size input to the operator.

        Box size used when merging by distance.
        default value is 1e-10.

        Parameters
        ----------
        my_box_size : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> op.inputs.box_size.connect(my_box_size)
        >>> # or
        >>> op.inputs.box_size(my_box_size)
        """
        return self._box_size

    @property
    def remove_duplicate_elements(self):
        """Allows to connect remove_duplicate_elements input to the operator.

        0: keep duplicate elements (default), 1:
        remove duplicate elements

        Parameters
        ----------
        my_remove_duplicate_elements : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> op.inputs.remove_duplicate_elements.connect(my_remove_duplicate_elements)
        >>> # or
        >>> op.inputs.remove_duplicate_elements(my_remove_duplicate_elements)
        """
        return self._remove_duplicate_elements


class OutputsMergeMeshes(_Outputs):
    """Intermediate class used to get outputs from
    merge_meshes operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_meshes()
    >>> # Connect inputs : op.inputs. ...
    >>> result_merges_mesh = op.outputs.merges_mesh()
    """

    def __init__(self, op: Operator):
        super().__init__(merge_meshes._spec().outputs, op)
        self._merges_mesh = Output(merge_meshes._spec().output_pin(0), 0, op)
        self._outputs.append(self._merges_mesh)

    @property
    def merges_mesh(self):
        """Allows to get merges_mesh output of the operator

        Returns
        ----------
        my_merges_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_meshes()
        >>> # Connect inputs : op.inputs. ...
        >>> result_merges_mesh = op.outputs.merges_mesh()
        """  # noqa: E501
        return self._merges_mesh



"""
on_reduced_coordinates
======================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class on_reduced_coordinates(Operator):
    """Evaluates a result on specified reduced coordinates of given elements
    (interpolates results inside elements with shape functions).

    Parameters
    ----------
    fields_container : FieldsContainer
    reduced_coordinates : Field or FieldsContainer
        Coordinates in the reference elements to find
        (found with the operator
        "find_reduced_coordinates")
    element_ids : ScopingsContainer
        Ids of the elements where each set of reduced
        coordinates is found (found with the
        operator "find_reduced_coordinates")
    mesh : MeshedRegion or MeshesContainer, optional
        If the first field in input has no mesh in
        support, then the mesh in this pin is
        expected (default is false), if a
        meshes container with several meshes
        is set, it should be on the same
        label spaces as the coordinates
        fields container
    use_quadratic_elements : bool
        If this pin is set to true interpolation is
        computed on the quadratic element if
        the element is quadratic (default is
        false). to use only when results have
        mid side nodes values.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mapping.on_reduced_coordinates()

    >>> # Make input connections
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_reduced_coordinates = dpf.Field()
    >>> op.inputs.reduced_coordinates.connect(my_reduced_coordinates)
    >>> my_element_ids = dpf.ScopingsContainer()
    >>> op.inputs.element_ids.connect(my_element_ids)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_use_quadratic_elements = bool()
    >>> op.inputs.use_quadratic_elements.connect(my_use_quadratic_elements)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mapping.on_reduced_coordinates(
    ...     fields_container=my_fields_container,
    ...     reduced_coordinates=my_reduced_coordinates,
    ...     element_ids=my_element_ids,
    ...     mesh=my_mesh,
    ...     use_quadratic_elements=my_use_quadratic_elements,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        fields_container=None,
        reduced_coordinates=None,
        element_ids=None,
        mesh=None,
        use_quadratic_elements=None,
        config=None,
        server=None,
    ):
        super().__init__(name="interpolation_operator", config=config, server=server)
        self._inputs = InputsOnReducedCoordinates(self)
        self._outputs = OutputsOnReducedCoordinates(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if reduced_coordinates is not None:
            self.inputs.reduced_coordinates.connect(reduced_coordinates)
        if element_ids is not None:
            self.inputs.element_ids.connect(element_ids)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if use_quadratic_elements is not None:
            self.inputs.use_quadratic_elements.connect(use_quadratic_elements)

    @staticmethod
    def _spec():
        description = """Evaluates a result on specified reduced coordinates of given elements
            (interpolates results inside elements with shape
            functions)."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="reduced_coordinates",
                    type_names=["field", "fields_container"],
                    optional=False,
                    document="""Coordinates in the reference elements to find
        (found with the operator
        "find_reduced_coordinates")""",
                ),
                2: PinSpecification(
                    name="element_ids",
                    type_names=["scopings_container"],
                    optional=False,
                    document="""Ids of the elements where each set of reduced
        coordinates is found (found with the
        operator "find_reduced_coordinates")""",
                ),
                7: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region", "meshes_container"],
                    optional=True,
                    document="""If the first field in input has no mesh in
        support, then the mesh in this pin is
        expected (default is false), if a
        meshes container with several meshes
        is set, it should be on the same
        label spaces as the coordinates
        fields container""",
                ),
                200: PinSpecification(
                    name="use_quadratic_elements",
                    type_names=["bool"],
                    optional=False,
                    document="""If this pin is set to true interpolation is
        computed on the quadratic element if
        the element is quadratic (default is
        false). to use only when results have
        mid side nodes values.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="interpolation_operator", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsOnReducedCoordinates
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsOnReducedCoordinates
        """
        return super().outputs


class InputsOnReducedCoordinates(_Inputs):
    """Intermediate class used to connect user inputs to
    on_reduced_coordinates operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mapping.on_reduced_coordinates()
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_reduced_coordinates = dpf.Field()
    >>> op.inputs.reduced_coordinates.connect(my_reduced_coordinates)
    >>> my_element_ids = dpf.ScopingsContainer()
    >>> op.inputs.element_ids.connect(my_element_ids)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_use_quadratic_elements = bool()
    >>> op.inputs.use_quadratic_elements.connect(my_use_quadratic_elements)
    """

    def __init__(self, op: Operator):
        super().__init__(on_reduced_coordinates._spec().inputs, op)
        self._fields_container = Input(
            on_reduced_coordinates._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_container)
        self._reduced_coordinates = Input(
            on_reduced_coordinates._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._reduced_coordinates)
        self._element_ids = Input(
            on_reduced_coordinates._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._element_ids)
        self._mesh = Input(on_reduced_coordinates._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._mesh)
        self._use_quadratic_elements = Input(
            on_reduced_coordinates._spec().input_pin(200), 200, op, -1
        )
        self._inputs.append(self._use_quadratic_elements)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def reduced_coordinates(self):
        """Allows to connect reduced_coordinates input to the operator.

        Coordinates in the reference elements to find
        (found with the operator
        "find_reduced_coordinates")

        Parameters
        ----------
        my_reduced_coordinates : Field or FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> op.inputs.reduced_coordinates.connect(my_reduced_coordinates)
        >>> # or
        >>> op.inputs.reduced_coordinates(my_reduced_coordinates)
        """
        return self._reduced_coordinates

    @property
    def element_ids(self):
        """Allows to connect element_ids input to the operator.

        Ids of the elements where each set of reduced
        coordinates is found (found with the
        operator "find_reduced_coordinates")

        Parameters
        ----------
        my_element_ids : ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> op.inputs.element_ids.connect(my_element_ids)
        >>> # or
        >>> op.inputs.element_ids(my_element_ids)
        """
        return self._element_ids

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        If the first field in input has no mesh in
        support, then the mesh in this pin is
        expected (default is false), if a
        meshes container with several meshes
        is set, it should be on the same
        label spaces as the coordinates
        fields container

        Parameters
        ----------
        my_mesh : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def use_quadratic_elements(self):
        """Allows to connect use_quadratic_elements input to the operator.

        If this pin is set to true interpolation is
        computed on the quadratic element if
        the element is quadratic (default is
        false). to use only when results have
        mid side nodes values.

        Parameters
        ----------
        my_use_quadratic_elements : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> op.inputs.use_quadratic_elements.connect(my_use_quadratic_elements)
        >>> # or
        >>> op.inputs.use_quadratic_elements(my_use_quadratic_elements)
        """
        return self._use_quadratic_elements


class OutputsOnReducedCoordinates(_Outputs):
    """Intermediate class used to get outputs from
    on_reduced_coordinates operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mapping.on_reduced_coordinates()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(on_reduced_coordinates._spec().outputs, op)
        self._fields_container = Output(
            on_reduced_coordinates._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.on_reduced_coordinates()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container



from .assemble_scalars_to_matrices import assemble_scalars_to_matrices 
from .assemble_scalars_to_matrices_fc import assemble_scalars_to_matrices_fc 
from .assemble_scalars_to_vectors import assemble_scalars_to_vectors 
from .assemble_scalars_to_vectors_fc import assemble_scalars_to_vectors_fc 
from .bind_support import bind_support 
from .bind_support_fc import bind_support_fc 
from .change_location import change_location 
from .change_shell_layers import change_shell_layers 
from .compute_time_scoping import compute_time_scoping 
from .default_value import default_value 
from .delegate_to_operator import delegate_to_operator 
from .ds_get_attribute import ds_get_attribute 
from .extract_field import extract_field 
from .extract_scoping import extract_scoping 
from .extract_sub_fc import extract_sub_fc 
from .extract_sub_mc import extract_sub_mc 
from .extract_sub_sc import extract_sub_sc 
from .extract_time_freq import extract_time_freq 
from .fc_get_attribute import fc_get_attribute 
from .field import field 
from .field_to_fc import field_to_fc 
from .fields_container import fields_container 
from .for_each import for_each 
from .forward import forward 
from .forward_field import forward_field 
from .forward_fields_container import forward_fields_container 
from .forward_meshes_container import forward_meshes_container 
from .hdf5dpf_workglow_provider import hdf5dpf_workglow_provider 
from .html_doc import html_doc 
from .incremental_concatenate_as_fc import incremental_concatenate_as_fc 
from .ints_to_scoping import ints_to_scoping 
from .make_for_each_range import make_for_each_range 
from .make_label_space import make_label_space 
from .make_overall import make_overall 
from .make_producer_consumer_for_each_iterator import make_producer_consumer_for_each_iterator 
from .merge_any import merge_any 
from .merge_data_tree import merge_data_tree 
from .merge_fields import merge_fields 
from .merge_fields_by_label import merge_fields_by_label 
from .merge_fields_containers import merge_fields_containers 
from .merge_generic_data_container import merge_generic_data_container 
from .merge_materials import merge_materials 
from .merge_meshes import merge_meshes 
from .merge_meshes_containers import merge_meshes_containers 
from .merge_property_fields import merge_property_fields 
from .merge_result_infos import merge_result_infos 
from .merge_scopings import merge_scopings 
from .merge_scopings_containers import merge_scopings_containers 
from .merge_supports import merge_supports 
from .merge_time_freq_supports import merge_time_freq_supports 
from .merge_weighted_fields import merge_weighted_fields 
from .merge_weighted_fields_containers import merge_weighted_fields_containers 
from .mesh import mesh 
from .overlap_fields import overlap_fields 
from .producer_consumer_for_each import producer_consumer_for_each 
from .property_field import property_field 
from .python_generator import python_generator 
from .python_script_exec import python_script_exec 
from .remote_operator_instantiate import remote_operator_instantiate 
from .remote_workflow_instantiate import remote_workflow_instantiate 
from .remove_unnecessary_labels import remove_unnecessary_labels 
from .scalars_to_field import scalars_to_field 
from .server_path import server_path 
from .set_attribute import set_attribute 
from .set_property import set_property 
from .split_in_for_each_range import split_in_for_each_range 
from .strain_from_voigt import strain_from_voigt 
from .txt_file_to_dpf import txt_file_to_dpf 
from .unitary_field import unitary_field 
from .weighted_merge_fields_by_label import weighted_merge_fields_by_label 



"""
fft_approx
==========
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class fft_approx(Operator):
    """Computes the fitting curve using FFT filtering and cubic fitting in
    space (node i: x=time, y=data), with the possibility to compute
    the first and the second derivatives of the curve.

    Parameters
    ----------
    time_scoping : Scoping, optional
        A time scoping to rescope / split the fields
        container given as input.
    mesh_scoping : Scoping or ScopingsContainer, optional
        A space (mesh entities) scoping (or scopings
        container) to rescope / split the
        fields container given as input.
    entity_to_fit : FieldsContainer
        Data changing in time to be fitted.
    component_number : int
        Component number as an integer, for example
        '0' for x-displacement, '1' for
        y-displacement, and so on.
    first_derivative : bool
        Calculate the first derivative (bool). the
        default is false.
    second_derivative : bool
        Calculate the second derivative (bool). the
        default is false.
    fit_data : bool
        Calculate the fitted values (bool). the
        default is false
    cutoff_fr : float or int, optional
        Cutoff frequency.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.math.fft_approx()

    >>> # Make input connections
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_entity_to_fit = dpf.FieldsContainer()
    >>> op.inputs.entity_to_fit.connect(my_entity_to_fit)
    >>> my_component_number = int()
    >>> op.inputs.component_number.connect(my_component_number)
    >>> my_first_derivative = bool()
    >>> op.inputs.first_derivative.connect(my_first_derivative)
    >>> my_second_derivative = bool()
    >>> op.inputs.second_derivative.connect(my_second_derivative)
    >>> my_fit_data = bool()
    >>> op.inputs.fit_data.connect(my_fit_data)
    >>> my_cutoff_fr = float()
    >>> op.inputs.cutoff_fr.connect(my_cutoff_fr)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.math.fft_approx(
    ...     time_scoping=my_time_scoping,
    ...     mesh_scoping=my_mesh_scoping,
    ...     entity_to_fit=my_entity_to_fit,
    ...     component_number=my_component_number,
    ...     first_derivative=my_first_derivative,
    ...     second_derivative=my_second_derivative,
    ...     fit_data=my_fit_data,
    ...     cutoff_fr=my_cutoff_fr,
    ... )

    >>> # Get output data
    >>> result_fitted_entity_y = op.outputs.fitted_entity_y()
    >>> result_first_der_dy = op.outputs.first_der_dy()
    >>> result_second_der_d2y = op.outputs.second_der_d2y()
    """

    def __init__(
        self,
        time_scoping=None,
        mesh_scoping=None,
        entity_to_fit=None,
        component_number=None,
        first_derivative=None,
        second_derivative=None,
        fit_data=None,
        cutoff_fr=None,
        config=None,
        server=None,
    ):
        super().__init__(name="fft_approx", config=config, server=server)
        self._inputs = InputsFftApprox(self)
        self._outputs = OutputsFftApprox(self)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if entity_to_fit is not None:
            self.inputs.entity_to_fit.connect(entity_to_fit)
        if component_number is not None:
            self.inputs.component_number.connect(component_number)
        if first_derivative is not None:
            self.inputs.first_derivative.connect(first_derivative)
        if second_derivative is not None:
            self.inputs.second_derivative.connect(second_derivative)
        if fit_data is not None:
            self.inputs.fit_data.connect(fit_data)
        if cutoff_fr is not None:
            self.inputs.cutoff_fr.connect(cutoff_fr)

    @staticmethod
    def _spec():
        description = """Computes the fitting curve using FFT filtering and cubic fitting in
            space (node i: x=time, y=data), with the possibility to
            compute the first and the second derivatives of the curve."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_scoping",
                    type_names=["vector<int32>", "scoping"],
                    optional=True,
                    document="""A time scoping to rescope / split the fields
        container given as input.""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["umap<int32,int32>", "scoping", "scopings_container"],
                    optional=True,
                    document="""A space (mesh entities) scoping (or scopings
        container) to rescope / split the
        fields container given as input.""",
                ),
                2: PinSpecification(
                    name="entity_to_fit",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Data changing in time to be fitted.""",
                ),
                3: PinSpecification(
                    name="component_number",
                    type_names=["int32"],
                    optional=False,
                    document="""Component number as an integer, for example
        '0' for x-displacement, '1' for
        y-displacement, and so on.""",
                ),
                4: PinSpecification(
                    name="first_derivative",
                    type_names=["bool"],
                    optional=False,
                    document="""Calculate the first derivative (bool). the
        default is false.""",
                ),
                5: PinSpecification(
                    name="second_derivative",
                    type_names=["bool"],
                    optional=False,
                    document="""Calculate the second derivative (bool). the
        default is false.""",
                ),
                6: PinSpecification(
                    name="fit_data",
                    type_names=["bool"],
                    optional=False,
                    document="""Calculate the fitted values (bool). the
        default is false""",
                ),
                7: PinSpecification(
                    name="cutoff_fr",
                    type_names=["double", "int32"],
                    optional=True,
                    document="""Cutoff frequency.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fitted_entity_y",
                    type_names=["fields_container"],
                    optional=False,
                    document="""The fitted entity is fitted using fft along
        the space scoping (node i: x=time,
        y=data). fitted y is expected to be
        close to the input data.""",
                ),
                1: PinSpecification(
                    name="first_der_dy",
                    type_names=["fields_container"],
                    optional=False,
                    document="""The first derivative (dy) from the fitted y.""",
                ),
                2: PinSpecification(
                    name="second_der_d2y",
                    type_names=["fields_container"],
                    optional=False,
                    document="""The second derivative (d2y) from the fitted
        y.""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="fft_approx", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsFftApprox
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsFftApprox
        """
        return super().outputs


class InputsFftApprox(_Inputs):
    """Intermediate class used to connect user inputs to
    fft_approx operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.fft_approx()
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_entity_to_fit = dpf.FieldsContainer()
    >>> op.inputs.entity_to_fit.connect(my_entity_to_fit)
    >>> my_component_number = int()
    >>> op.inputs.component_number.connect(my_component_number)
    >>> my_first_derivative = bool()
    >>> op.inputs.first_derivative.connect(my_first_derivative)
    >>> my_second_derivative = bool()
    >>> op.inputs.second_derivative.connect(my_second_derivative)
    >>> my_fit_data = bool()
    >>> op.inputs.fit_data.connect(my_fit_data)
    >>> my_cutoff_fr = float()
    >>> op.inputs.cutoff_fr.connect(my_cutoff_fr)
    """

    def __init__(self, op: Operator):
        super().__init__(fft_approx._spec().inputs, op)
        self._time_scoping = Input(fft_approx._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._time_scoping)
        self._mesh_scoping = Input(fft_approx._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._entity_to_fit = Input(fft_approx._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._entity_to_fit)
        self._component_number = Input(fft_approx._spec().input_pin(3), 3, op, -1)
        self._inputs.append(self._component_number)
        self._first_derivative = Input(fft_approx._spec().input_pin(4), 4, op, -1)
        self._inputs.append(self._first_derivative)
        self._second_derivative = Input(fft_approx._spec().input_pin(5), 5, op, -1)
        self._inputs.append(self._second_derivative)
        self._fit_data = Input(fft_approx._spec().input_pin(6), 6, op, -1)
        self._inputs.append(self._fit_data)
        self._cutoff_fr = Input(fft_approx._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._cutoff_fr)

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        A time scoping to rescope / split the fields
        container given as input.

        Parameters
        ----------
        my_time_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        A space (mesh entities) scoping (or scopings
        container) to rescope / split the
        fields container given as input.

        Parameters
        ----------
        my_mesh_scoping : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def entity_to_fit(self):
        """Allows to connect entity_to_fit input to the operator.

        Data changing in time to be fitted.

        Parameters
        ----------
        my_entity_to_fit : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.entity_to_fit.connect(my_entity_to_fit)
        >>> # or
        >>> op.inputs.entity_to_fit(my_entity_to_fit)
        """
        return self._entity_to_fit

    @property
    def component_number(self):
        """Allows to connect component_number input to the operator.

        Component number as an integer, for example
        '0' for x-displacement, '1' for
        y-displacement, and so on.

        Parameters
        ----------
        my_component_number : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.component_number.connect(my_component_number)
        >>> # or
        >>> op.inputs.component_number(my_component_number)
        """
        return self._component_number

    @property
    def first_derivative(self):
        """Allows to connect first_derivative input to the operator.

        Calculate the first derivative (bool). the
        default is false.

        Parameters
        ----------
        my_first_derivative : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.first_derivative.connect(my_first_derivative)
        >>> # or
        >>> op.inputs.first_derivative(my_first_derivative)
        """
        return self._first_derivative

    @property
    def second_derivative(self):
        """Allows to connect second_derivative input to the operator.

        Calculate the second derivative (bool). the
        default is false.

        Parameters
        ----------
        my_second_derivative : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.second_derivative.connect(my_second_derivative)
        >>> # or
        >>> op.inputs.second_derivative(my_second_derivative)
        """
        return self._second_derivative

    @property
    def fit_data(self):
        """Allows to connect fit_data input to the operator.

        Calculate the fitted values (bool). the
        default is false

        Parameters
        ----------
        my_fit_data : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.fit_data.connect(my_fit_data)
        >>> # or
        >>> op.inputs.fit_data(my_fit_data)
        """
        return self._fit_data

    @property
    def cutoff_fr(self):
        """Allows to connect cutoff_fr input to the operator.

        Cutoff frequency.

        Parameters
        ----------
        my_cutoff_fr : float or int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> op.inputs.cutoff_fr.connect(my_cutoff_fr)
        >>> # or
        >>> op.inputs.cutoff_fr(my_cutoff_fr)
        """
        return self._cutoff_fr


class OutputsFftApprox(_Outputs):
    """Intermediate class used to get outputs from
    fft_approx operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.fft_approx()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fitted_entity_y = op.outputs.fitted_entity_y()
    >>> result_first_der_dy = op.outputs.first_der_dy()
    >>> result_second_der_d2y = op.outputs.second_der_d2y()
    """

    def __init__(self, op: Operator):
        super().__init__(fft_approx._spec().outputs, op)
        self._fitted_entity_y = Output(fft_approx._spec().output_pin(0), 0, op)
        self._outputs.append(self._fitted_entity_y)
        self._first_der_dy = Output(fft_approx._spec().output_pin(1), 1, op)
        self._outputs.append(self._first_der_dy)
        self._second_der_d2y = Output(fft_approx._spec().output_pin(2), 2, op)
        self._outputs.append(self._second_der_d2y)

    @property
    def fitted_entity_y(self):
        """Allows to get fitted_entity_y output of the operator

        Returns
        ----------
        my_fitted_entity_y : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fitted_entity_y = op.outputs.fitted_entity_y()
        """  # noqa: E501
        return self._fitted_entity_y

    @property
    def first_der_dy(self):
        """Allows to get first_der_dy output of the operator

        Returns
        ----------
        my_first_der_dy : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> # Connect inputs : op.inputs. ...
        >>> result_first_der_dy = op.outputs.first_der_dy()
        """  # noqa: E501
        return self._first_der_dy

    @property
    def second_der_d2y(self):
        """Allows to get second_der_d2y output of the operator

        Returns
        ----------
        my_second_der_d2y : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.fft_approx()
        >>> # Connect inputs : op.inputs. ...
        >>> result_second_der_d2y = op.outputs.second_der_d2y()
        """  # noqa: E501
        return self._second_der_d2y



"""
Session
========
"""
import abc
import ctypes
import logging
import threading
import traceback
import warnings
import weakref

from ansys.dpf.gate import session_capi, session_grpcapi, capi

from ansys.dpf.core import server as server_module
from ansys.dpf.core import server_types, errors
from ansys.dpf.core.check_version import version_requires, server_meet_version
from ansys.dpf.core.common import (
    _common_percentage_progress_bar,
    _progress_bar_is_available,
)
from ansys.dpf.core import data_tree


LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")


@capi.GenericCallBackType
def progress_call_back(obj, nature, arg):
    try:
        obj = ctypes.cast(obj, ctypes.POINTER(ctypes.py_object))
        handler = obj.contents.value
        if nature == 0:
            handler.started_operators += 1
        elif nature == 1:
            handler.finished_operators += 1
            if handler.finished_operators > 0 and handler.bar:
                handler.bar.update(handler.finished_operators / handler.started_operators * 100)
                if handler.finished_operators == handler.started_operators:
                    handler.bar.finish()
        elif nature == 9:
            handler.finished_operators = 0
            handler.started_operators = 0
        elif handler.bar:
            handler.bar.update(0)
    except Exception as e:
        print(e.args)


class EventHandlerBase:
    @abc.abstractmethod
    def add_operator(self, operator, pin, identifier):
        pass

    @abc.abstractmethod
    def start_listening(self):
        pass


class EventHandler(EventHandlerBase):
    def __init__(self, session):
        self._session = weakref.ref(session)
        self.bar = None
        self.started_operators = 0
        self.finished_operators = 0
        self.py_obj = ctypes.py_object(self)
        self._session()._api.add_external_event_handler(
            self._session(),
            ctypes.cast(ctypes.pointer(self.py_obj), ctypes.c_void_p),
            progress_call_back,
        )

    def start_listening(self):
        if not _progress_bar_is_available():
            print("Progress bar is not available, please install progressbar2")
            return
        self.bar = _common_percentage_progress_bar("Workflow running")
        self.started_operators = 0
        self.finished_operators = 0

    def add_operator(self, operator, pin, identifier):
        from ansys.dpf.core import workflow

        wf = workflow.Workflow(server=self._session()._server)
        wf.add_operator(operator)
        wf.set_output_name("out", operator, pin)
        wf._api.work_flow_discover_operators(wf)
        self._session().add_workflow(wf, identifier)


class GrpcEventHandler(EventHandlerBase):
    def __init__(self, session):
        self._session = weakref.ref(session)
        self.bar = None
        self._session()._api.add_external_event_handler(self._session(), self, None)

    def start_listening(self):
        if not _progress_bar_is_available():
            print("Progress bar is not available, please install progressbar2")
            return
        self.bar = _common_percentage_progress_bar("Workflow running")
        thread = threading.Thread(
            target=self._session()._api.start_listening,
            args=[self._session(), self.bar, LOG],
        )
        return thread

    def add_operator(self, operator, pin, identifier):
        self._session()._api.add_operator(self._session(), identifier, operator, pin)


class Session:
    """A class used to create a user session on the server, it allows to plan events
    call backs from the server: progress bar when workflows are running, logging...
    A session is started every time a ``'DpfServer'`` is created.

    Notes
    -----
    Class available with server's version starting at 3.0.
    """

    def __init__(self, server=None):
        # step 1: get server
        server = server_module.get_or_create_server(server)
        if not server.meet_version("3.0"):
            raise errors.DpfVersionNotSupported("3.0")
        self._server_weak_ref = weakref.ref(server)
        # step 2: get api
        self._api = server.get_api_for_type(
            capi=session_capi.SessionCAPI, grpcapi=session_grpcapi.SessionGRPCAPI
        )

        # step3: init environment
        self._api.init_session_environment(self)  # creates stub when gRPC

        if self._server.has_client():
            self._internal_obj = self._api.session_new_on_client(self._server.client)
        else:
            self._internal_obj = self._api.session_new()

        self._handler = None
        if server_meet_version("3.0", server):
            self.add_progress_system()

        self._released = False

    @property
    def _server(self):
        return self._server_weak_ref()

    @version_requires("3.0")
    def add_workflow(self, workflow, identifier):
        """Add a workflow to the session. It allows to follow the workflow's
        events while it's running.
        This method is automatically called when a workflow's output
        is requested.

        Parameters
        ----------
        workflow : Workflow

        identifier : str
            name given to the workflow
        """
        if self._handler is not None:
            self._api.add_workflow(self, identifier, workflow)

    @version_requires("3.0")
    def add_operator(self, operator, pin, identifier):
        """Creates a workflow made of the input operator and all its ancestors
        to the session. It allows to follow the workflow's
        events while it's running.
        This method is automatically called when an operator's output
        is requested and the property :func:`ansys.dpf.core.dpf_operator.Operator.progress_bar`
        is set to ``'True'``.

        Parameters
        ----------
        operator : Operator

        pin : int
            output pin number requested

        identifier : str
            name given to the workflow
        """
        if self._handler is not None:
            self._handler.add_operator(operator, pin, identifier)

    @version_requires("6.1")
    def handle_events_with_file_logger(self, file_path, verbosity_level=1):
        """Adds an event handler of type ``file_logger`` server side.
        Events will then be caught and forwarded to the file stream.

        Parameters
        ----------
        file_path : str

        verbosity_level : int
            0, 1 or 2
        """
        properties = data_tree.DataTree(server=self._server)
        properties.add({"file_name": file_path, "verbosity": verbosity_level})
        self._api.add_event_handler_type(self, "file_logger", properties)

    @version_requires("6.1")
    def start_emitting_rpc_log(self):
        """Adds a signal emitter to the session. This emitter will catch all incoming rpc calls.
        Adding a handler will enable the logging (
        use :func:`Session.handle_events_with_file_logger()`).

        """
        self._api.add_signal_emitter_type(
            self, "gRPC_calls_interceptor", "gRPC_calls_interceptor", None
        )

    @version_requires("3.0")
    def listen_to_progress(self):
        """Starts a progress bar and updates it every time an operator is
        finished.
        """
        if self._handler is not None:
            return self._handler.start_listening()

    def _init_handler(self):
        if isinstance(self._server, server_types.InProcessServer):
            self._handler = EventHandler(self)
        elif isinstance(self._server, server_types.LegacyGrpcServer):
            self._handler = GrpcEventHandler(self)

    @version_requires("3.0")
    def add_progress_system(self):
        """Asks the session to start recording progress events.
        Called when the session is started.
        """
        self._init_handler()

    @version_requires("3.0")
    def flush_workflows(self):
        """This removes the handle on the workflow by the ``session``"""
        self._api.flush_workflows(self)

    def delete(self):
        try:
            if not self._released:
                self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())
        self._released = True

    def __del__(self):
        self.delete()



"""
Core
====
"""
import os
import logging
import warnings
import weakref

from ansys.dpf.core import errors, misc
from ansys.dpf.core import server as server_module
from ansys.dpf.core.check_version import version_requires
from ansys.dpf.core.runtime_config import (
    RuntimeClientConfig,
    RuntimeCoreConfig,
)
from ansys.dpf.gate import (
    data_processing_capi,
    data_processing_grpcapi,
    tmp_dir_capi,
    tmp_dir_grpcapi,
    collection_capi,
    collection_grpcapi,
    integral_types,
    object_handler,
)

try:
    from grpc import _channel  # noqa: F401

    # weirdly necessary to delete LegacyGrpcError
except ImportError:
    pass

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")

if "DPF_CONFIGURATION" in os.environ:
    CONFIGURATION = os.environ["DPF_CONFIGURATION"]
else:
    CONFIGURATION = "release"


def load_library(filename, name="", symbol="LoadOperators", server=None, generate_operators=False):
    """Dynamically load an operators library for dpf.core.
    Code containing this library's operators is generated in
    ansys.dpf.core.operators

    Parameters
    ----------
    filename : str or os.PathLike
        Filename of the operator library.

    name : str, optional
        Library name.  Probably optional

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    generate_operators : bool, optional
        Whether operators code generation should be done or not (default is False).

    Examples
    --------
    Load the mesh operators for Windows (for Linux, just use
    'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

    >>> from ansys.dpf import core as dpf
    >>> # dpf.load_library('meshOperatorsCore.dll', 'mesh_operators')

    """
    base = BaseService(server, load_operators=False)
    base.load_library(filename, name, symbol, generate_operators)
    return name + " successfully loaded"


def upload_file_in_tmp_folder(file_path, new_file_name=None, server=None):
    """Upload a file from the client to the server in a temporary folder
    deleted when the server is shutdown

    Parameters
    ----------
    file_path : str or os.PathLike
        file path on the client side to upload

    new_file_name : str, optional
        name to give to the file server side,
        if no name is specified, the same name as the input file is given

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
      server_file_path : str
           path generated server side

    Notes
    -----
    Is not implemented for usage with type(server)=
    :class:`ansys.dpf.core.server_types.InProcessServer`.
    """
    base = BaseService(server, load_operators=False)
    return base.upload_file_in_tmp_folder(file_path, new_file_name)


def upload_files_in_folder(
    to_server_folder_path, client_folder_path, specific_extension=None, server=None
):
    """Upload all the files from a folder of the client
    to the target server folder path.

    Parameters
    ----------
    to_server_folder_path : str or os.PathLike
        folder path target where will be uploaded files on the server side

    client_folder_path: str or os.PathLike
        folder path where the files that must be uploaded are located
        on client side

    specific_extension (optional) : str
        copies only the files with the given extension

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
    paths : list of str
        new file paths server side
    """
    base = BaseService(server, load_operators=False)
    return base.upload_files_in_folder(
        to_server_folder_path, client_folder_path, specific_extension
    )


def download_file(server_file_path, to_client_file_path, server=None):
    """Download a file from the server to the target client file path

    Parameters
    ----------
    server_file_path : str or os.PathLike
        file path to download on the server side

    to_client_file_path: str or os.PathLike
        file path target where the file will be located client side

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Notes
    -----
    Is not implemented for usage with type(server)=
    :class:`ansys.dpf.core.server_types.InProcessServer`.
    """
    base = BaseService(server, load_operators=False)
    return base.download_file(server_file_path, to_client_file_path)


def download_files_in_folder(
    server_folder_path, to_client_folder_path, specific_extension=None, server=None
):
    """Download all the files from a folder of the server
    to the target client folder path

    Parameters
    ----------
    server_folder_path : str or os.PathLike
        folder path to download on the server side

    to_client_folder_path: str or os.PathLike
        folder path target where the files will be located client side

    specific_extension (optional) : str
        copies only the files with the given extension

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
    paths : list of str
        new file paths client side

    """
    base = BaseService(server, load_operators=False)
    return base.download_files_in_folder(
        server_folder_path, to_client_folder_path, specific_extension
    )


def upload_file(file_path, to_server_file_path, server=None):
    """Upload a file from the client to the target server file path

    Parameters
    ----------
    file_path : str or os.PathLike
        file path on the client side to upload

    to_server_file_path: str or os.PathLike
        file path target where the file will be located server side

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
    server_file_path : str
        path generated server side
    """
    base = BaseService(server, load_operators=False)
    return base.upload_file(file_path, to_server_file_path)


def make_tmp_dir_server(server=None):
    """Create a temporary folder server side. Only one temporary folder can be created
    by server instance.
    The folder will be deleted when the server is stopped.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
    path : str
        path to the temporary dir
    """
    base = BaseService(server, load_operators=False)
    return base.make_tmp_dir_server()


def _description(dpf_entity_message, server=None):
    """Ask the server to describe the entity in input

    Parameters
    ----------
    dpf_entity_message : core.Operator._message, core.Workflow._message,
                         core.Scoping._message, core.Field._message,
                         core.FieldContainer._message, core.MeshedRegion._message

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
       description : str
    """
    try:
        return BaseService(server, load_operators=False)._description(dpf_entity_message)
    except:
        return ""


def _deep_copy(dpf_entity, server=None):
    """Returns a copy of the entity in the requested server

    Parameters
    ----------
    dpf_entity: core.Operator, core.Workflow, core.Scoping,
                core.Field, core.FieldsContainer, core.MeshedRegion...
        Dpf entity to deep_copy

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    Returns
    -------
       deep_copy of dpf_entity: core.Operator, core.Workflow, core.Scoping,
                                core.Field, core.FieldsContainer, core.MeshedRegion...
    """
    from ansys.dpf.core.operators.serialization import serializer_to_string, string_deserializer
    from ansys.dpf.core.common import types_enum_to_types

    serializer = serializer_to_string(server=server)
    serializer.connect(1, dpf_entity)
    deserializer = string_deserializer(server=server)
    deserializer.connect(0, serializer, 0)
    type_map = types_enum_to_types()
    output_type = list(type_map.keys())[list(type_map.values()).index(dpf_entity.__class__)]
    return deserializer.get_output(1, output_type)


class BaseService:
    """The Base Service class allows to make generic requests to dpf's server.
    For example, information about the server can be requested,
    uploading/downloading file from and to the server can be done,
    new operators plugins can be loaded...
    Most of the request done by the BaseService class are wrapped by
    functions.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the global server.

    timeout : float, optional
        Fails when a connection takes longer than ``timeout`` seconds
        to initialize.

    load_operators : bool, optional
        Automatically load the math operators

    Examples
    --------
    Connect to an existing DPF server
    >>> from ansys.dpf import core as dpf
    >>> #server = dpf.connect_to_server(ip='127.0.0.1', port = 50054, as_global=False)
    >>> #base = dpf.BaseService(server=server)

    """

    def __init__(self, server=None, load_operators=True, timeout=5):
        """Initialize base service"""
        # step 1: get server
        if server is None:
            server = server_module.get_or_create_server(server)
        self._server = weakref.ref(server)
        self._collection_api = None

        # step 2: get api
        self._api = self._server().get_api_for_type(
            capi=data_processing_capi.DataProcessingCAPI,
            grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
        )
        self._api_tmp_dir = self._server().get_api_for_type(
            capi=tmp_dir_capi.TmpDirCAPI, grpcapi=tmp_dir_grpcapi.TmpDirGRPCAPI
        )

        # step3: init environment
        self._api.init_data_processing_environment(self)  # creates stub when gRPC

    def make_tmp_dir_server(self):
        """Create a temporary folder server side. Only one temporary folder can be created
        by server instance.
        The folder will be deleted when the server is stopped.

        Returns
        -------
        path : str
            path to the temporary dir
        """
        if self._server().has_client():
            return self._api_tmp_dir.tmp_dir_get_dir_on_client(client=self._server().client)
        else:
            return self._api_tmp_dir.tmp_dir_get_dir()

    def load_library(self, file_path, name="", symbol="LoadOperators", generate_operators=False):
        """Dynamically load an operators library for dpf.core.
        Code containing this library's operators is generated in
        ansys.dpf.core.operators

        Parameters
        ----------
        file_path : str or os.PathLike
            file_path of the operator library.

        name : str, optional
            Library name.  Probably optional

        generate_operators : bool, optional
            Whether operators code generation should be done or not (default is False).

        Examples
        --------
        Load the mesh operators for Windows (for Linux, just use
        'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

        >>> from ansys.dpf import core as dpf
        >>> base = dpf.BaseService()
        >>> # base.load_library('meshOperatorsCore.dll', 'mesh_operators')

        """
        file_path = str(file_path)
        if self._server().has_client():
            self._internal_obj = self._api.data_processing_load_library_on_client(
                sLibraryKey=name,
                sDllPath=file_path,
                sloader_symbol=symbol,
                client=self._server().client,
            )
        else:
            self._internal_obj = self._api.data_processing_load_library(
                name=name, dllPath=file_path, symbol=symbol
            )
        if generate_operators:
            # TODO: fix code generation upload posix
            import os

            def __generate_code(TARGET_PATH, filename, name, symbol):
                from ansys.dpf.core.dpf_operator import Operator

                try:
                    code_gen = Operator("python_generator")
                    code_gen.connect(1, TARGET_PATH)
                    code_gen.connect(0, filename)
                    code_gen.connect(2, symbol)
                    code_gen.connect(3, name)
                    code_gen.run()
                except Exception as e:
                    warnings.warn("Unable to generate the python code with error: " + str(e.args))

            local_dir = os.path.dirname(os.path.abspath(__file__))
            LOCAL_PATH = os.path.join(local_dir, "operators")
            if not self._server().local_server:
                if self._server().os != "posix" or (not self._server().os and os.name != "posix"):
                    # send local generated code
                    TARGET_PATH = self.make_tmp_dir_server()
                    self.upload_files_in_folder(TARGET_PATH, LOCAL_PATH, "py")

                    # generate code
                    __generate_code(TARGET_PATH, file_path, name, symbol)

                    try:
                        self.download_files_in_folder(TARGET_PATH, LOCAL_PATH, "py")
                    except Exception as e:
                        warnings.warn(
                            f"Unable to download the python generated code with error: {e.args}"
                        )
            else:
                __generate_code(
                    TARGET_PATH=LOCAL_PATH, filename=file_path, name=name, symbol=symbol
                )

    @version_requires("6.0")
    def apply_context(self, context):
        """Defines the settings that will be used to load DPF's plugins.
        A DPF xml file can be used to list the plugins and set up variables.

        Parameters
        ----------
        context : ServerContext
            The context allows to choose which capabilities are available server side.

        Notes
        -----
        Available with server's version starting at 6.0 (Ansys 2023R2).
        """
        if not self._server().meet_version("6.0"):
            raise errors.DpfVersionNotSupported("6.0")
        if self._server().has_client():
            self._api.data_processing_apply_context_on_client(
                self._server().client,
                int(context.licensing_context_type),
                context.xml_path,
            )
        else:
            self._api.data_processing_apply_context(
                int(context.licensing_context_type), context.xml_path
            )

    def initialize_with_context(self, context):
        """Defines the settings that will be used to initialize DPF.
        A DPF xml file can be used to list the plugins and set up variables.

        Parameters
        ----------
        context : ServerContext
            The context allows to choose which capabilities are available server side.

        Notes
        -----
        Available with server's version starting at 4.0 (Ansys 2022R2) for InProcess Server
        and starting at 6.0 (Ansys 2023R2) for Grpc Servers.
        """
        if self._server().has_client():
            if not self._server().meet_version("6.0"):
                raise errors.DpfVersionNotSupported("6.0")
            self._api.data_processing_initialize_with_context_on_client(
                self._server().client,
                int(context.licensing_context_type),
                context.xml_path,
            )
        else:
            if not self._server().meet_version("4.0"):
                raise errors.DpfVersionNotSupported("4.0")
            self._api.data_processing_initialize_with_context(
                int(context.licensing_context_type), context.xml_path
            )

    @version_requires("6.0")
    def release_dpf(self):
        """Clears the available Operators and Releases licenses when necessary.

        Notes
        -----
        Available with server's version starting at 6.0 (Ansys 2023R2).
        """
        if self._server().has_client():
            error = self._api.data_processing_release_on_client(self._server().client, 1)
        else:
            error = self._api.data_processing_release(1)

    def get_runtime_client_config(self):
        if self._server().has_client():
            data_tree_tmp = self._api.data_processing_get_client_config_as_data_tree()
            config_to_return = RuntimeClientConfig(data_tree=data_tree_tmp, server=self._server())
        else:
            if misc.RUNTIME_CLIENT_CONFIG is None:
                from ansys.dpf.core import data_tree

                misc.RUNTIME_CLIENT_CONFIG = RuntimeClientConfig(
                    data_tree=data_tree.DataTree(server=self._server())
                )
                misc.RUNTIME_CLIENT_CONFIG._data_tree._holds_server = False
            config_to_return = misc.RUNTIME_CLIENT_CONFIG
        return config_to_return

    @version_requires("4.0")
    def get_runtime_core_config(self):
        if self._server().has_client():
            data_tree_tmp = self._api.data_processing_get_global_config_as_data_tree_on_client(
                self._server().client
            )

        else:
            data_tree_tmp = self._api.data_processing_get_global_config_as_data_tree()
        return RuntimeCoreConfig(data_tree=data_tree_tmp, server=self._server())

    @property
    def server_info(self):
        """Send the request for server information and keep
        the info into a dictionary

        Returns
        -------
        info : dictionary
            dictionary with "server_ip", "server_port", "server_process_id"
            "server_version" keys
        """
        return self._get_server_info()

    def _get_server_info(self):
        serv_ip = ""
        serv_port = integral_types.MutableInt32(-1)
        proc_id = ""
        serv_ver_maj = integral_types.MutableInt32(-1)
        serv_ver_min = integral_types.MutableInt32(-1)
        serv_os = ""
        # ip/port
        if self._server().has_client():
            serv_ip = self._api.data_processing_get_server_ip_and_port(
                client=self._server().client, port=serv_port
            )
            serv_port = int(serv_port)
        else:
            serv_ip = ""
            serv_port = None
        # process id
        if self._server().has_client():
            proc_id = self._api.data_processing_process_id_on_client(client=self._server().client)
        else:
            proc_id = self._api.data_processing_process_id()
        # server version
        if self._server().has_client():
            self._api.data_processing_get_server_version_on_client(
                client=self._server().client, major=serv_ver_maj, minor=serv_ver_min
            )
        else:
            self._api.data_processing_get_server_version(major=serv_ver_maj, minor=serv_ver_min)
        # server os
        if self._server().has_client():
            serv_os = self._api.data_processing_get_os_on_client(client=self._server().client)
        else:
            serv_os = self._api.data_processing_get_os()

        out = {
            "server_ip": serv_ip,
            "server_port": serv_port,
            "server_process_id": proc_id,
            "server_version": str(int(serv_ver_maj)) + "." + str(int(serv_ver_min)),
            "os": serv_os,
        }

        return out

    def _description(self, dpf_entity_message):
        """Ask the server to describe the entity in input

        Parameters
        ----------
        dpf_entity_message : core.Operator._message, core.Workflow._message,
                             core.Scoping._message, core.Field._message,
        core.FieldContainer._message, core.MeshedRegion._message...

        Returns
        -------
           description : str
        """
        data = object_handler.ObjHandler(
            data_processing_api=self._api,
            internal_obj=dpf_entity_message,
            server=self._server(),
        )
        data.get_ownership()
        return self._api.data_processing_description_string(data=data)

    def _get_separator(self, path):
        s1 = len(path.split("\\"))
        s2 = len(path.split("/"))
        if s2 > s1:
            # Linux case
            separator = "/"
        elif s1 > s2:
            # Windows case
            separator = "\\"
        return separator

    def download_file(self, server_file_path, to_client_file_path):
        """Download a file from the server to the target client file path

        Parameters
        ----------
        server_file_path : str or os.PathLike
            file path to download on the server side

        to_client_file_path: str or os.PathLike
            file path target where the file will be located client side
        """
        if not self._server().has_client():
            txt = """
            download service only available for server with gRPC communication protocol
            """
            raise errors.ServerTypeError(txt)
        client_path = self._api.data_processing_download_file(
            client=self._server().client,
            server_file_path=str(server_file_path),
            to_client_file_path=str(to_client_file_path),
        )

    def _set_collection_api(self):
        if self._collection_api is None:
            self._collection_api = self._server().get_api_for_type(
                capi=collection_capi.CollectionCAPI,
                grpcapi=collection_grpcapi.CollectionGRPCAPI,
            )
            self._collection_api.init_collection_environment(self)
        return self._collection_api

    def download_files_in_folder(
        self, server_folder_path, to_client_folder_path, specific_extension=None
    ):
        """Download all the files from a folder of the server
        to the target client folder path

        Parameters
        ----------
        server_folder_path : str or os.PathLike
            folder path to download on the server side

        to_client_folder_path: str or os.PathLike
            folder path target where the files will be located client side

        specific_extension (optional) : str
            copies only the files with the given extension

        Returns
        -------
        paths : list of str
            new file paths client side

        """
        if not self._server().has_client():
            txt = """
            download service only available for server with gRPC communication protocol
            """
            raise ValueError(txt)
        if specific_extension is None:
            specific_extension = ""
        client_paths_ptr = self._api.data_processing_download_files(
            client=self._server().client,
            server_file_path=str(server_folder_path),
            to_client_file_path=str(to_client_folder_path),
            specific_extension=specific_extension,
        )
        if not isinstance(client_paths_ptr, list):
            from ansys.dpf.gate import object_handler

            # collection of string
            client_paths = object_handler.ObjHandler(
                data_processing_api=self._api,
                internal_obj=client_paths_ptr,
                server=self._server(),
            )
            coll_api = self._set_collection_api()
            size = coll_api.collection_get_size(client_paths)
            out = [0] * size
            for i in range(0, size):
                entry = coll_api.collection_get_string_entry(client_paths, i)
                out[i] = entry
            return out
        return client_paths_ptr

    def upload_files_in_folder(
        self, to_server_folder_path, client_folder_path, specific_extension=None
    ):
        """Upload all the files from a folder of the client
        to the target server folder path.

        Parameters
        ----------
        to_server_folder_path : str or os.PathLike
            folder path target where will be uploaded files on the server side

        client_folder_path: str or os.PathLike
            folder path where the files that must be uploaded are located
            on client side

        specific_extension (optional) : str
            copies only the files with the given extension

        Returns
        -------
        paths : list of str
            new file paths server side
        """
        server_paths = []
        for root, subdirectories, files in os.walk(client_folder_path):
            for subdirectory in subdirectories:
                subdir = os.path.join(root, subdirectory)
                for filename in os.listdir(subdir):
                    f = os.path.join(subdir, filename)
                    server_paths = self._upload_and_get_server_path(
                        specific_extension,
                        f,
                        filename,
                        server_paths,
                        str(to_server_folder_path),
                        subdirectory,
                    )
            for file in files:
                f = os.path.join(root, file)
                server_paths = self._upload_and_get_server_path(
                    specific_extension,
                    f,
                    file,
                    server_paths,
                    str(to_server_folder_path),
                )
            break
        return server_paths

    def _upload_and_get_server_path(
        self,
        specific_extension,
        f,
        filename,
        server_paths,
        to_server_folder_path,
        subdirectory=None,
    ):
        separator = self._get_separator(to_server_folder_path)

        if subdirectory is not None:
            to_server_file_path = (
                to_server_folder_path + separator + subdirectory + separator + filename
            )
        else:
            to_server_file_path = to_server_folder_path + separator + filename
        if ((specific_extension is not None) and (f.endswith(specific_extension))) or (
            specific_extension is None
        ):
            if not self._server().has_client():
                txt = """
                download service only available for server with gRPC communication protocol
                """
                raise errors.ServerTypeError(txt)
            server_path = self._api.data_processing_upload_file(
                client=self._server().client,
                file_path=f,
                to_server_file_path=to_server_file_path,
                use_tmp_dir=False,
            )
            server_paths.append(server_path)
        return server_paths

    def upload_file(self, file_path, to_server_file_path):
        """Upload a file from the client to the target server file path

        Parameters
        ----------
        file_path : str or os.PathLike
            file path on the client side to upload

        to_server_file_path: str or os.PathLike
            file path target where the file will be located server side

        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if os.stat(file_path).st_size == 0:
            raise ValueError(file_path + " is empty")
        if not self._server().has_client():
            txt = """
            download service only available for server with gRPC communication protocol
            """
            raise errors.ServerTypeError(txt)
        return self._api.data_processing_upload_file(
            client=self._server().client,
            file_path=str(file_path),
            to_server_file_path=str(to_server_file_path),
            use_tmp_dir=False,
        )

    def upload_file_in_tmp_folder(self, file_path, new_file_name=None):
        """Upload a file from the client to the server in a temporary folder
        deleted when the server is shutdown

        Parameters
        ----------
        file_path : str or os.PathLike
            file path on the client side to upload

        new_file_name : str, optional
            name to give to the file server side,
            if no name is specified, the same name as the input file is given

        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if new_file_name:
            file_name = new_file_name
        else:
            file_name = os.path.basename(file_path)
        if os.stat(file_path).st_size == 0:
            raise ValueError(file_path + " is empty")
        if not self._server().has_client():
            txt = """
            download service only available for server with gRPC communication protocol
            """
            raise errors.ServerTypeError(txt)
        return self._api.data_processing_upload_file(
            client=self._server().client,
            file_path=str(file_path),
            to_server_file_path=str(file_name),
            use_tmp_dir=True,
        )

    def _prepare_shutdown(self):
        if self._server().has_client():
            self._api.data_processing_prepare_shutdown(client=self._server().client)

    # @version_requires("4.0")
    def _release_server(self):
        """
        Release the reference taken by this client on the server

        Notes
        -----
        Should be used only if the server was started by this client's instance.
        To use only with server version > 4.0
        """
        if self._server().has_client():
            self._api.data_processing_release_server(client=self._server().client)



"""
change_shell_layers
===================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class change_shell_layers(Operator):
    """Extract the expected shell layers from the input fields. If the fields
    contain only one layer and the permissive configuration input is
    set to true then it returns the input fields. If permissive
    configuration input is set to false, any change which should not
    be permitted won't be achieved and corresponding field in output
    will be empty. If permissive configuration input is set to true
    (default), carefully check the result.

    Parameters
    ----------
    fields_container : FieldsContainer or Field
    e_shell_layer : int
        0:top, 1: bottom, 2: bottomtop, 3:mid,
        4:bottomtopmid
    mesh : MeshedRegion or MeshesContainer, optional
        Mesh support of the input fields_container,
        in case it does not have one defined.
        if the fields_container contains
        mixed shell/solid results, the mesh
        is required (either by connecting
        this pin or in the support).


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.change_shell_layers()

    >>> # Make input connections
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_e_shell_layer = int()
    >>> op.inputs.e_shell_layer.connect(my_e_shell_layer)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.change_shell_layers(
    ...     fields_container=my_fields_container,
    ...     e_shell_layer=my_e_shell_layer,
    ...     mesh=my_mesh,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        fields_container=None,
        e_shell_layer=None,
        mesh=None,
        config=None,
        server=None,
    ):
        super().__init__(name="change_shellLayers", config=config, server=server)
        self._inputs = InputsChangeShellLayers(self)
        self._outputs = OutputsChangeShellLayers(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if e_shell_layer is not None:
            self.inputs.e_shell_layer.connect(e_shell_layer)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        description = """Extract the expected shell layers from the input fields. If the fields
            contain only one layer and the permissive configuration
            input is set to true then it returns the input fields. If
            permissive configuration input is set to false, any change
            which should not be permitted won't be achieved and
            corresponding field in output will be empty. If permissive
            configuration input is set to true (default), carefully
            check the result."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container", "field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="e_shell_layer",
                    type_names=["int32", "enum dataProcessing::EShellLayers"],
                    optional=False,
                    document="""0:top, 1: bottom, 2: bottomtop, 3:mid,
        4:bottomtopmid""",
                ),
                2: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region", "meshes_container"],
                    optional=True,
                    document="""Mesh support of the input fields_container,
        in case it does not have one defined.
        if the fields_container contains
        mixed shell/solid results, the mesh
        is required (either by connecting
        this pin or in the support).""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container", "field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="change_shellLayers", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsChangeShellLayers
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsChangeShellLayers
        """
        return super().outputs


class InputsChangeShellLayers(_Inputs):
    """Intermediate class used to connect user inputs to
    change_shell_layers operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.change_shell_layers()
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_e_shell_layer = int()
    >>> op.inputs.e_shell_layer.connect(my_e_shell_layer)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    """

    def __init__(self, op: Operator):
        super().__init__(change_shell_layers._spec().inputs, op)
        self._fields_container = Input(
            change_shell_layers._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_container)
        self._e_shell_layer = Input(change_shell_layers._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._e_shell_layer)
        self._mesh = Input(change_shell_layers._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._mesh)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Parameters
        ----------
        my_fields_container : FieldsContainer or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.change_shell_layers()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def e_shell_layer(self):
        """Allows to connect e_shell_layer input to the operator.

        0:top, 1: bottom, 2: bottomtop, 3:mid,
        4:bottomtopmid

        Parameters
        ----------
        my_e_shell_layer : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.change_shell_layers()
        >>> op.inputs.e_shell_layer.connect(my_e_shell_layer)
        >>> # or
        >>> op.inputs.e_shell_layer(my_e_shell_layer)
        """
        return self._e_shell_layer

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Mesh support of the input fields_container,
        in case it does not have one defined.
        if the fields_container contains
        mixed shell/solid results, the mesh
        is required (either by connecting
        this pin or in the support).

        Parameters
        ----------
        my_mesh : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.change_shell_layers()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh


class OutputsChangeShellLayers(_Outputs):
    """Intermediate class used to get outputs from
    change_shell_layers operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.change_shell_layers()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(change_shell_layers._spec().outputs, op)
        self.fields_container_as_fields_container = Output(
            _modify_output_spec_with_one_type(
                change_shell_layers._spec().output_pin(0), "fields_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.fields_container_as_fields_container)
        self.fields_container_as_field = Output(
            _modify_output_spec_with_one_type(
                change_shell_layers._spec().output_pin(0), "field"
            ),
            0,
            op,
        )
        self._outputs.append(self.fields_container_as_field)



"""
time_freq_interpolation
=======================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class time_freq_interpolation(Operator):
    """Interpolates between all the matching fields of a fields container at
    given times or frequencies, using ramped: fieldOut =
    field1*(1.-fact)+field2*(fact), or stepped: fieldOut=field2. If
    the time freq is higher than the max available, the field at the
    max time freq is taken. Computes the output time freq support to
    support the fields container

    Parameters
    ----------
    fields_container : FieldsContainer
    time_freq_values : float or Field
        List of frequencies or times needed. to
        specify load steps, put a field (and
        not a list) in input with a scoping
        located on "timefreq_steps".
    step : int, optional
        If a field is set as input, the step ids
        should be its scoping.
    interpolation_type : int, optional
        1 is ramped, 2 is stepped, default is 1.
    force_new_time_freq_support : bool, optional
        If set to true, the output fields container
        will always have a new time freq
        support rescoped to the output
        time_freq_values (default is false).
        if set to false, the time freq
        support is only recreated when time
        or frequency values are between
        existing ones.
    time_freq_support : TimeFreqSupport, optional


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.math.time_freq_interpolation()

    >>> # Make input connections
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_time_freq_values = float()
    >>> op.inputs.time_freq_values.connect(my_time_freq_values)
    >>> my_step = int()
    >>> op.inputs.step.connect(my_step)
    >>> my_interpolation_type = int()
    >>> op.inputs.interpolation_type.connect(my_interpolation_type)
    >>> my_force_new_time_freq_support = bool()
    >>> op.inputs.force_new_time_freq_support.connect(my_force_new_time_freq_support)
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.math.time_freq_interpolation(
    ...     fields_container=my_fields_container,
    ...     time_freq_values=my_time_freq_values,
    ...     step=my_step,
    ...     interpolation_type=my_interpolation_type,
    ...     force_new_time_freq_support=my_force_new_time_freq_support,
    ...     time_freq_support=my_time_freq_support,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        fields_container=None,
        time_freq_values=None,
        step=None,
        interpolation_type=None,
        force_new_time_freq_support=None,
        time_freq_support=None,
        config=None,
        server=None,
    ):
        super().__init__(name="time_freq_interpolation", config=config, server=server)
        self._inputs = InputsTimeFreqInterpolation(self)
        self._outputs = OutputsTimeFreqInterpolation(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if time_freq_values is not None:
            self.inputs.time_freq_values.connect(time_freq_values)
        if step is not None:
            self.inputs.step.connect(step)
        if interpolation_type is not None:
            self.inputs.interpolation_type.connect(interpolation_type)
        if force_new_time_freq_support is not None:
            self.inputs.force_new_time_freq_support.connect(force_new_time_freq_support)
        if time_freq_support is not None:
            self.inputs.time_freq_support.connect(time_freq_support)

    @staticmethod
    def _spec():
        description = """Interpolates between all the matching fields of a fields container at
            given times or frequencies, using ramped: fieldOut =
            field1*(1.-fact)+field2*(fact), or stepped:
            fieldOut=field2. If the time freq is higher than the max
            available, the field at the max time freq is taken.
            Computes the output time freq support to support the
            fields container"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="time_freq_values",
                    type_names=["double", "vector<double>", "field"],
                    optional=False,
                    document="""List of frequencies or times needed. to
        specify load steps, put a field (and
        not a list) in input with a scoping
        located on "timefreq_steps".""",
                ),
                2: PinSpecification(
                    name="step",
                    type_names=["int32"],
                    optional=True,
                    document="""If a field is set as input, the step ids
        should be its scoping.""",
                ),
                3: PinSpecification(
                    name="interpolation_type",
                    type_names=["int32"],
                    optional=True,
                    document="""1 is ramped, 2 is stepped, default is 1.""",
                ),
                4: PinSpecification(
                    name="force_new_time_freq_support",
                    type_names=["bool"],
                    optional=True,
                    document="""If set to true, the output fields container
        will always have a new time freq
        support rescoped to the output
        time_freq_values (default is false).
        if set to false, the time freq
        support is only recreated when time
        or frequency values are between
        existing ones.""",
                ),
                8: PinSpecification(
                    name="time_freq_support",
                    type_names=["time_freq_support"],
                    optional=True,
                    document="""""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="time_freq_interpolation", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsTimeFreqInterpolation
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsTimeFreqInterpolation
        """
        return super().outputs


class InputsTimeFreqInterpolation(_Inputs):
    """Intermediate class used to connect user inputs to
    time_freq_interpolation operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.time_freq_interpolation()
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_time_freq_values = float()
    >>> op.inputs.time_freq_values.connect(my_time_freq_values)
    >>> my_step = int()
    >>> op.inputs.step.connect(my_step)
    >>> my_interpolation_type = int()
    >>> op.inputs.interpolation_type.connect(my_interpolation_type)
    >>> my_force_new_time_freq_support = bool()
    >>> op.inputs.force_new_time_freq_support.connect(my_force_new_time_freq_support)
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)
    """

    def __init__(self, op: Operator):
        super().__init__(time_freq_interpolation._spec().inputs, op)
        self._fields_container = Input(
            time_freq_interpolation._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_container)
        self._time_freq_values = Input(
            time_freq_interpolation._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._time_freq_values)
        self._step = Input(time_freq_interpolation._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._step)
        self._interpolation_type = Input(
            time_freq_interpolation._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._interpolation_type)
        self._force_new_time_freq_support = Input(
            time_freq_interpolation._spec().input_pin(4), 4, op, -1
        )
        self._inputs.append(self._force_new_time_freq_support)
        self._time_freq_support = Input(
            time_freq_interpolation._spec().input_pin(8), 8, op, -1
        )
        self._inputs.append(self._time_freq_support)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def time_freq_values(self):
        """Allows to connect time_freq_values input to the operator.

        List of frequencies or times needed. to
        specify load steps, put a field (and
        not a list) in input with a scoping
        located on "timefreq_steps".

        Parameters
        ----------
        my_time_freq_values : float or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.time_freq_values.connect(my_time_freq_values)
        >>> # or
        >>> op.inputs.time_freq_values(my_time_freq_values)
        """
        return self._time_freq_values

    @property
    def step(self):
        """Allows to connect step input to the operator.

        If a field is set as input, the step ids
        should be its scoping.

        Parameters
        ----------
        my_step : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.step.connect(my_step)
        >>> # or
        >>> op.inputs.step(my_step)
        """
        return self._step

    @property
    def interpolation_type(self):
        """Allows to connect interpolation_type input to the operator.

        1 is ramped, 2 is stepped, default is 1.

        Parameters
        ----------
        my_interpolation_type : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.interpolation_type.connect(my_interpolation_type)
        >>> # or
        >>> op.inputs.interpolation_type(my_interpolation_type)
        """
        return self._interpolation_type

    @property
    def force_new_time_freq_support(self):
        """Allows to connect force_new_time_freq_support input to the operator.

        If set to true, the output fields container
        will always have a new time freq
        support rescoped to the output
        time_freq_values (default is false).
        if set to false, the time freq
        support is only recreated when time
        or frequency values are between
        existing ones.

        Parameters
        ----------
        my_force_new_time_freq_support : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.force_new_time_freq_support.connect(my_force_new_time_freq_support)
        >>> # or
        >>> op.inputs.force_new_time_freq_support(my_force_new_time_freq_support)
        """
        return self._force_new_time_freq_support

    @property
    def time_freq_support(self):
        """Allows to connect time_freq_support input to the operator.

        Parameters
        ----------
        my_time_freq_support : TimeFreqSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> op.inputs.time_freq_support.connect(my_time_freq_support)
        >>> # or
        >>> op.inputs.time_freq_support(my_time_freq_support)
        """
        return self._time_freq_support


class OutputsTimeFreqInterpolation(_Outputs):
    """Intermediate class used to get outputs from
    time_freq_interpolation operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.time_freq_interpolation()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(time_freq_interpolation._spec().outputs, op)
        self._fields_container = Output(
            time_freq_interpolation._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.time_freq_interpolation()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container
